import torch
from torch import nn
from torch.nn import Linear, Embedding
from torch_geometric.nn.inits import glorot_orthogonal
from torch_geometric.nn import radius_graph
from torch_scatter import scatter
from scipy import special as sp
from math import sqrt
from torch_sparse import SparseTensor
from torch_geometric.nn import knn_graph
import torch.nn.functional as F
import numpy as np
import sympy as sym
from scipy.optimize import brentq
from math import sqrt, pi as PI
from torch.nn import ReLU, Sequential, BatchNorm1d
from torch.nn import Dropout, Linear


def Jn(r, n):
    return np.sqrt(np.pi / (2 * r)) * sp.jv(n + 0.5, r)


def Jn_zeros(n, k):
    zerosj = np.zeros((n, k), dtype='float32')
    zerosj[0] = np.arange(1, k + 1) * np.pi
    points = np.arange(1, k + n) * np.pi
    racines = np.zeros(k + n - 1, dtype='float32')
    for i in range(1, n):
        for j in range(k + n - 1 - i):
            foo = brentq(Jn, points[j], points[j + 1], (i, ))
            racines[j] = foo
        points = racines
        zerosj[i][:k] = racines[:k]

    return zerosj


def spherical_bessel_formulas(n):
    x = sym.symbols('x')

    f = [sym.sin(x) / x]
    a = sym.sin(x) / x
    for i in range(1, n):
        b = sym.diff(a, x) / x
        f += [sym.simplify(b * (-x)**i)]
        a = sym.simplify(b)
    return f




def bessel_basis(n, k):
    zeros = Jn_zeros(n, k)
    normalizer = []
    for order in range(n):
        normalizer_tmp = []
        for i in range(k):
            normalizer_tmp += [0.5 * Jn(zeros[order, i], order + 1)**2]
        normalizer_tmp = 1 / np.array(normalizer_tmp)**0.5
        normalizer += [normalizer_tmp]

    f = spherical_bessel_formulas(n)
    x = sym.symbols('x')
    bess_basis = []
    for order in range(n):
        bess_basis_tmp = []
        for i in range(k):
            bess_basis_tmp += [
                sym.simplify(normalizer[order][i] *
                             f[order].subs(x, zeros[order, i] * x))
            ]
        bess_basis += [bess_basis_tmp]
    return bess_basis




def sph_harm_prefactor(k, m):
    return ((2 * k + 1) * np.math.factorial(k - abs(m)) /
            (4 * np.pi * np.math.factorial(k + abs(m))))**0.5


def associated_legendre_polynomials(k, zero_m_only=True):
    z = sym.symbols('z')
    P_l_m = [[0] * (j + 1) for j in range(k)]

    P_l_m[0][0] = 1
    if k > 0:
        P_l_m[1][0] = z

        for j in range(2, k):
            P_l_m[j][0] = sym.simplify(((2 * j - 1) * z * P_l_m[j - 1][0] -
                                        (j - 1) * P_l_m[j - 2][0]) / j)
        if not zero_m_only:
            for i in range(1, k):
                P_l_m[i][i] = sym.simplify((1 - 2 * i) * P_l_m[i - 1][i - 1])
                if i + 1 < k:
                    P_l_m[i + 1][i] = sym.simplify(
                        (2 * i + 1) * z * P_l_m[i][i])
                for j in range(i + 2, k):
                    P_l_m[j][i] = sym.simplify(
                        ((2 * j - 1) * z * P_l_m[j - 1][i] -
                         (i + j - 1) * P_l_m[j - 2][i]) / (j - i))

    return P_l_m


def real_sph_harm(l, zero_m_only=False, spherical_coordinates=True):
    """
    Computes formula strings of the the real part of the spherical harmonics up to order l (excluded).
    Variables are either cartesian coordinates x,y,z on the unit sphere or spherical coordinates phi and theta.
    """
    if not zero_m_only:
        x = sym.symbols('x')
        y = sym.symbols('y')
        S_m = [x*0]
        C_m = [1+0*x]
        # S_m = [0]
        # C_m = [1]
        for i in range(1, l):
            x = sym.symbols('x')
            y = sym.symbols('y')
            S_m += [x*S_m[i-1] + y*C_m[i-1]]
            C_m += [x*C_m[i-1] - y*S_m[i-1]]

    P_l_m = associated_legendre_polynomials(l, zero_m_only)
    if spherical_coordinates:
        theta = sym.symbols('theta')
        z = sym.symbols('z')
        for i in range(len(P_l_m)):
            for j in range(len(P_l_m[i])):
                if type(P_l_m[i][j]) != int:
                    P_l_m[i][j] = P_l_m[i][j].subs(z, sym.cos(theta))
        if not zero_m_only:
            phi = sym.symbols('phi')
            for i in range(len(S_m)):
                S_m[i] = S_m[i].subs(x, sym.sin(
                    theta)*sym.cos(phi)).subs(y, sym.sin(theta)*sym.sin(phi))
            for i in range(len(C_m)):
                C_m[i] = C_m[i].subs(x, sym.sin(
                    theta)*sym.cos(phi)).subs(y, sym.sin(theta)*sym.sin(phi))

    Y_func_l_m = [['0']*(2*j + 1) for j in range(l)]
    for i in range(l):
        Y_func_l_m[i][0] = sym.simplify(sph_harm_prefactor(i, 0) * P_l_m[i][0])

    if not zero_m_only:
        for i in range(1, l):
            for j in range(1, i + 1):
                Y_func_l_m[i][j] = sym.simplify(
                    2**0.5 * sph_harm_prefactor(i, j) * C_m[j] * P_l_m[i][j])
        for i in range(1, l):
            for j in range(1, i + 1):
                Y_func_l_m[i][-j] = sym.simplify(
                    2**0.5 * sph_harm_prefactor(i, -j) * S_m[j] * P_l_m[i][j])

    return Y_func_l_m


def get_nearst_node(pos, batch):
    j, i = knn_graph(pos, 1, batch)
    adj_nearest = SparseTensor(row=i, col=j, value=torch.ones_like(j, device=j.device), sparse_sizes=(batch.size(0), batch.size(0)))
    j2, i2 = knn_graph(pos, 2, batch)
    adj_nearest2 = SparseTensor(row=i2, col=j2, value=torch.ones_like(j2, device=j2.device), sparse_sizes=(batch.size(0), batch.size(0)))

    return adj_nearest, SparseTensor.from_dense(adj_nearest2.to_dense() - adj_nearest.to_dense())



def xyztodat(pos, edge_index, num_nodes, batch):
    j, i = edge_index  # j->i

    # Calculate distances. # number of edges
    dist = (pos[i] - pos[j]).pow(2).sum(dim=-1).sqrt()

    value = torch.arange(j.size(0), device=j.device)
    adj_t = SparseTensor(row=i, col=j, value=value, sparse_sizes=(num_nodes, num_nodes))
    adj_t_row = adj_t[j]
    num_triplets = adj_t_row.set_value(None).sum(dim=1).to(torch.long)

    # Node indices (k->j->i) for triplets.
    idx_i = i.repeat_interleave(num_triplets)
    idx_j = j.repeat_interleave(num_triplets)
    idx_k = adj_t_row.storage.col()
    mask = idx_i != idx_k
    idx_i, idx_j, idx_k = idx_i[mask], idx_j[mask], idx_k[mask]

    # Edge indices (k-j, j->i) for triplets.
    idx_kj = adj_t_row.storage.value()[mask]
    idx_ji = adj_t_row.storage.row()[mask]

    # Calculate angles. 0 to pi
    pos_ji = pos[idx_i] - pos[idx_j]
    pos_jk = pos[idx_k] - pos[idx_j]
    a = (pos_ji * pos_jk).sum(dim=-1) # cos_angle * |pos_ji| * |pos_jk|
    b = torch.cross(pos_ji, pos_jk).norm(dim=-1) # |sin_angle| * |pos_ji| * |pos_jk|
    angle = torch.atan2(b, a)

    adj_nearest, adj_nearest2 = get_nearst_node(pos, batch)
    
    adj_nearest_row = adj_nearest[idx_j]
    adj_nearest2_row = adj_nearest2[idx_j]
    idx_k_n = adj_nearest_row.storage.col()
    idx_k_n2 = adj_nearest2_row.storage.col()
    mask = idx_k_n == idx_i
    idx_k_n[mask] = idx_k_n2[mask]

    # Calculate torsions.
    pos_j0 = pos[idx_k] - pos[idx_j]
    pos_ji = pos[idx_i] - pos[idx_j]
    pos_jk = pos[idx_k_n] - pos[idx_j]
    dist_ji = pos_ji.pow(2).sum(dim=-1).sqrt()
    plane1 = torch.cross(pos_ji, pos_j0)
    plane2 = torch.cross(pos_ji, pos_jk)
    a = (plane1 * plane2).sum(dim=-1) # cos_angle * |plane1| * |plane2|
    b = (torch.cross(plane1, plane2) * pos_ji).sum(dim=-1) / dist_ji
    torsion = torch.atan2(b, a) # -pi to pi
    torsion[torsion<=0]+=2*PI # 0 to 2pi

    return dist, angle, torsion, i, j, idx_kj, idx_ji


class Envelope(torch.nn.Module):
    def __init__(self, exponent):
        super(Envelope, self).__init__()
        self.p = exponent + 1
        self.a = -(self.p + 1) * (self.p + 2) / 2
        self.b = self.p * (self.p + 2)
        self.c = -self.p * (self.p + 1) / 2

    def forward(self, x):
        p, a, b, c = self.p, self.a, self.b, self.c
        x_pow_p0 = x.pow(p - 1)
        x_pow_p1 = x_pow_p0 * x
        x_pow_p2 = x_pow_p1 * x
        return 1. / x + a * x_pow_p0 + b * x_pow_p1 + c * x_pow_p2



class dist_emb(torch.nn.Module):
    def __init__(self, num_radial, cutoff=6.0, envelope_exponent=5):
        super(dist_emb, self).__init__()
        self.cutoff = cutoff
        self.envelope = Envelope(envelope_exponent)

        self.freq = torch.nn.Parameter(torch.Tensor(num_radial))

        self.reset_parameters()

    def reset_parameters(self):
        # torch.arange(1, self.freq.numel() + 1, out=self.freq).mul_(PI)
        pass

    def forward(self, dist):
        dist = dist.unsqueeze(-1) / self.cutoff
        return self.envelope(dist) * (self.freq * dist).sin()


class angle_emb(torch.nn.Module):
    def __init__(self, num_spherical, num_radial, cutoff=6.0,
                 envelope_exponent=5):
        super(angle_emb, self).__init__()
        assert num_radial <= 64
        self.num_spherical = num_spherical
        self.num_radial = num_radial
        self.cutoff = cutoff
        # self.envelope = Envelope(envelope_exponent)

        bessel_forms = bessel_basis(num_spherical, num_radial)
        sph_harm_forms = real_sph_harm(num_spherical)
        self.sph_funcs = []
        self.bessel_funcs = []

        x, theta = sym.symbols('x theta')
        modules = {'sin': torch.sin, 'cos': torch.cos}
        for i in range(num_spherical):
            if i == 0:
                sph1 = sym.lambdify([theta], sph_harm_forms[i][0], modules)(0)
                self.sph_funcs.append(lambda x: torch.zeros_like(x) + sph1)
            else:
                sph = sym.lambdify([theta], sph_harm_forms[i][0], modules)
                self.sph_funcs.append(sph)
            for j in range(num_radial):
                bessel = sym.lambdify([x], bessel_forms[i][j], modules)
                self.bessel_funcs.append(bessel)

    def forward(self, dist, angle, idx_kj):
        dist = dist / self.cutoff
        rbf = torch.stack([f(dist) for f in self.bessel_funcs], dim=1)
        # rbf = self.envelope(dist).unsqueeze(-1) * rbf

        cbf = torch.stack([f(angle) for f in self.sph_funcs], dim=1)

        n, k = self.num_spherical, self.num_radial
        out = (rbf[idx_kj].view(-1, n, k) * cbf.view(-1, n, 1)).view(-1, n * k)
        return out


class torsion_emb(torch.nn.Module):
    def __init__(self, num_spherical, num_radial, cutoff=6.0,
                 envelope_exponent=5):
        super(torsion_emb, self).__init__()
        assert num_radial <= 64
        self.num_spherical = num_spherical #
        self.num_radial = num_radial
        self.cutoff = cutoff
        # self.envelope = Envelope(envelope_exponent)

        bessel_forms = bessel_basis(num_spherical, num_radial)
        sph_harm_forms = real_sph_harm(num_spherical, zero_m_only=False)
        self.sph_funcs = []
        self.bessel_funcs = []

        x = sym.symbols('x')
        theta = sym.symbols('theta')
        phi = sym.symbols('phi')
        modules = {'sin': torch.sin, 'cos': torch.cos}
        for i in range(self.num_spherical):
            if i == 0:
                sph1 = sym.lambdify([theta, phi], sph_harm_forms[i][0], modules)
                self.sph_funcs.append(lambda x, y: torch.zeros_like(x) + torch.zeros_like(y) + sph1(0,0)) #torch.zeros_like(x) + torch.zeros_like(y)
            else:
                for k in range(-i, i + 1):
                    sph = sym.lambdify([theta, phi], sph_harm_forms[i][k+i], modules)
                    self.sph_funcs.append(sph)
            for j in range(self.num_radial):
                bessel = sym.lambdify([x], bessel_forms[i][j], modules)
                self.bessel_funcs.append(bessel)

    def forward(self, dist, angle, phi, idx_kj):
        dist = dist / self.cutoff
        rbf = torch.stack([f(dist) for f in self.bessel_funcs], dim=1)
        cbf = torch.stack([f(angle, phi) for f in self.sph_funcs], dim=1)

        n, k = self.num_spherical, self.num_radial
        out = (rbf[idx_kj].view(-1, 1, n, k) * cbf.view(-1, n, n, 1)).view(-1, n * n * k)
        return out



def swish(x):
    return x * torch.sigmoid(x)


class emb(torch.nn.Module):
    def __init__(self, num_spherical, num_radial, cutoff, envelope_exponent):
        super(emb, self).__init__()
        self.dist_emb = dist_emb(num_radial, cutoff, envelope_exponent)
        self.angle_emb = angle_emb(num_spherical, num_radial, cutoff, envelope_exponent)
        self.torsion_emb = torsion_emb(num_spherical, num_radial, cutoff, envelope_exponent)
        self.reset_parameters()

    def reset_parameters(self):
        self.dist_emb.reset_parameters()

    def forward(self, dist, angle, torsion, idx_kj):
        dist_emb = self.dist_emb(dist)
        angle_emb = self.angle_emb(dist, angle, idx_kj)
        torsion_emb = self.torsion_emb(dist, angle, torsion, idx_kj)
        return dist_emb, angle_emb, torsion_emb

class ResidualLayer(torch.nn.Module):
    def __init__(self, hidden_channels, act=swish):
        super(ResidualLayer, self).__init__()
        self.act = act
        self.lin1 = Linear(hidden_channels, hidden_channels)
        self.lin2 = Linear(hidden_channels, hidden_channels)

        self.reset_parameters()

    def reset_parameters(self):
        glorot_orthogonal(self.lin1.weight, scale=2.0)
        self.lin1.bias.data.fill_(0)
        glorot_orthogonal(self.lin2.weight, scale=2.0)
        self.lin2.bias.data.fill_(0)

    def forward(self, x):
        return x + self.act(self.lin2(self.act(self.lin1(x))))

class FeatureEmbedding(torch.nn.Module):
    def __init__(self, input_dim, output_dim):
        super(FeatureEmbedding, self).__init__()
        self.network = Sequential(
            Linear(input_dim, output_dim // 2),
            BatchNorm1d(output_dim // 2),
            ReLU(),
            Linear(output_dim // 2, output_dim),
            ReLU()
        )
    
    def forward(self, x):
        return self.network(x)
    
    def reset_parameters(self):
        for module in self.network:
            if hasattr(module, 'reset_parameters'):
                module.reset_parameters()

class init(torch.nn.Module):
    def __init__(self, num_radial, hidden_channels, act=swish, one_hot=False):
        super(init, self).__init__()
        self.act = act
        self.one_hot = one_hot
        
        if self.one_hot:
            self.emb = Embedding(100, hidden_channels)
        else: # option to use no node features and a learned embedding vector for each node instead
            # self.node_embedding = nn.Parameter(torch.empty((hidden_channels,)))
            # nn.init.normal_(self.node_embedding)
            self.emb = FeatureEmbedding(28, hidden_channels)
        
        self.lin_rbf_0 = Linear(num_radial, hidden_channels)
        self.lin = Linear(3 * hidden_channels, hidden_channels)
        self.lin_rbf_1 = nn.Linear(num_radial, hidden_channels, bias=False)
        self.reset_parameters()

    def reset_parameters(self):
        self.emb.reset_parameters()
        self.lin_rbf_0.reset_parameters()
        self.lin.reset_parameters()
        glorot_orthogonal(self.lin_rbf_1.weight, scale=2.0)

    def forward(self, x, emb, i, j):
        rbf,_,_ = emb
        x = self.emb(x)

        rbf0 = self.act(self.lin_rbf_0(rbf))
        e1 = self.act(self.lin(torch.cat([x[i], x[j], rbf0], dim=-1)))
        e2 = self.lin_rbf_1(rbf) * e1

        return e1, e2


class update_e(torch.nn.Module):
    def __init__(self, hidden_channels, int_emb_size, basis_emb_size_dist, basis_emb_size_angle, basis_emb_size_torsion, num_spherical, num_radial,
        num_before_skip, num_after_skip, act=swish):
        super(update_e, self).__init__()
        self.act = act
        self.lin_rbf1 = nn.Linear(num_radial, basis_emb_size_dist, bias=False)
        self.lin_rbf2 = nn.Linear(basis_emb_size_dist, hidden_channels, bias=False)
        self.lin_sbf1 = nn.Linear(num_spherical * num_radial, basis_emb_size_angle, bias=False)
        self.lin_sbf2 = nn.Linear(basis_emb_size_angle, int_emb_size, bias=False)
        self.lin_t1 = nn.Linear(num_spherical * num_spherical * num_radial, basis_emb_size_torsion, bias=False)
        self.lin_t2 = nn.Linear(basis_emb_size_torsion, int_emb_size, bias=False)
        self.lin_rbf = nn.Linear(num_radial, hidden_channels, bias=False)

        self.lin_kj = nn.Linear(hidden_channels, hidden_channels)
        self.lin_ji = nn.Linear(hidden_channels, hidden_channels)

        self.lin_down = nn.Linear(hidden_channels, int_emb_size, bias=False)
        self.lin_up = nn.Linear(int_emb_size, hidden_channels, bias=False)

        self.layers_before_skip = torch.nn.ModuleList([
            ResidualLayer(hidden_channels, act)
            for _ in range(num_before_skip)
        ])
        self.lin = nn.Linear(hidden_channels, hidden_channels)
        self.layers_after_skip = torch.nn.ModuleList([
            ResidualLayer(hidden_channels, act)
            for _ in range(num_after_skip)
        ])

        self.reset_parameters()

    def reset_parameters(self):
        glorot_orthogonal(self.lin_rbf1.weight, scale=2.0)
        glorot_orthogonal(self.lin_rbf2.weight, scale=2.0)
        glorot_orthogonal(self.lin_sbf1.weight, scale=2.0)
        glorot_orthogonal(self.lin_sbf2.weight, scale=2.0)
        glorot_orthogonal(self.lin_t1.weight, scale=2.0)
        glorot_orthogonal(self.lin_t2.weight, scale=2.0)

        glorot_orthogonal(self.lin_kj.weight, scale=2.0)
        self.lin_kj.bias.data.fill_(0)
        glorot_orthogonal(self.lin_ji.weight, scale=2.0)
        self.lin_ji.bias.data.fill_(0)

        glorot_orthogonal(self.lin_down.weight, scale=2.0)
        glorot_orthogonal(self.lin_up.weight, scale=2.0)

        for res_layer in self.layers_before_skip:
            res_layer.reset_parameters()
        glorot_orthogonal(self.lin.weight, scale=2.0)
        self.lin.bias.data.fill_(0)
        for res_layer in self.layers_after_skip:
            res_layer.reset_parameters()

        glorot_orthogonal(self.lin_rbf.weight, scale=2.0)

    def forward(self, x, emb, idx_kj, idx_ji):
        rbf0, sbf, t = emb
        x1,_ = x

        x_ji = self.act(self.lin_ji(x1))
        x_kj = self.act(self.lin_kj(x1))

        rbf = self.lin_rbf1(rbf0)
        rbf = self.lin_rbf2(rbf)
        x_kj = x_kj * rbf

        x_kj = self.act(self.lin_down(x_kj))

        sbf = self.lin_sbf1(sbf)
        sbf = self.lin_sbf2(sbf)
        x_kj = x_kj[idx_kj] * sbf

        t = self.lin_t1(t)
        t = self.lin_t2(t)
        x_kj = x_kj * t

        x_kj = scatter(x_kj, idx_ji, dim=0, dim_size=x1.size(0))
        x_kj = self.act(self.lin_up(x_kj))

        e1 = x_ji + x_kj
        for layer in self.layers_before_skip:
            e1 = layer(e1)
        e1 = self.act(self.lin(e1)) + x1
        for layer in self.layers_after_skip:
            e1 = layer(e1)
        e2 = self.lin_rbf(rbf0) * e1

        return e1, e2


class update_v(torch.nn.Module):
    def __init__(self, hidden_channels, out_emb_channels, out_channels, num_output_layers, act, output_init):
        super(update_v, self).__init__()
        self.act = act
        self.output_init = output_init

        self.lin_up = nn.Linear(hidden_channels, out_emb_channels, bias=True)
        self.lins = torch.nn.ModuleList()
        for _ in range(num_output_layers):
            self.lins.append(nn.Linear(out_emb_channels, out_emb_channels))
        self.lin = nn.Linear(out_emb_channels, out_channels, bias=False)

        self.reset_parameters()

    def reset_parameters(self):
        glorot_orthogonal(self.lin_up.weight, scale=2.0)
        for lin in self.lins:
            glorot_orthogonal(lin.weight, scale=2.0)
            lin.bias.data.fill_(0)
        if self.output_init == 'zeros':
            self.lin.weight.data.fill_(0)
        if self.output_init == 'GlorotOrthogonal':
            glorot_orthogonal(self.lin.weight, scale=2.0)

    def forward(self, e, i):
        _, e2 = e
        v = scatter(e2, i, dim=0)
        v = self.lin_up(v)
        for lin in self.lins:
            v = self.act(lin(v))
        v = self.lin(v)
        return v


class update_u(torch.nn.Module):
    def __init__(self):
        super(update_u, self).__init__()

    def forward(self, u, v, batch):
        u += scatter(v, batch, dim=0)
        return u

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
def xyz_to_dat(pos, edge_index, num_nodes, use_torsion = False):
    """
    Compute the diatance, angle, and torsion from geometric information.

    Args:
        pos: Geometric information for every node in the graph.
        edge_index: Edge index of the graph.
        number_nodes: Number of nodes in the graph.
        use_torsion: If set to :obj:`True`, will return distance, angle and torsion, otherwise only return distance and angle (also retrun some useful index). (default: :obj:`False`)
    """
    j, i = edge_index  # j->i

    # Calculate distances. # number of edges
    dist = (pos[i] - pos[j]).pow(2).sum(dim=-1).sqrt()

    value = torch.arange(j.size(0), device=j.device)
    adj_t = SparseTensor(row=i, col=j, value=value, sparse_sizes=(num_nodes, num_nodes))
    adj_t_row = adj_t[j]
    num_triplets = adj_t_row.set_value(None).sum(dim=1).to(torch.long)

    # Node indices (k->j->i) for triplets.
    idx_i = i.repeat_interleave(num_triplets)
    idx_j = j.repeat_interleave(num_triplets)
    idx_k = adj_t_row.storage.col()
    mask = idx_i != idx_k
    idx_i, idx_j, idx_k = idx_i[mask], idx_j[mask], idx_k[mask]

    # Edge indices (k-j, j->i) for triplets.
    idx_kj = adj_t_row.storage.value()[mask]
    idx_ji = adj_t_row.storage.row()[mask]

    # Calculate angles. 0 to pi
    pos_ji = pos[idx_i] - pos[idx_j]
    pos_jk = pos[idx_k] - pos[idx_j]
    a = (pos_ji * pos_jk).sum(dim=-1) # cos_angle * |pos_ji| * |pos_jk|
    b = torch.cross(pos_ji, pos_jk).norm(dim=-1) # sin_angle * |pos_ji| * |pos_jk|
    angle = torch.atan2(b, a)


    if use_torsion:
        # Prepare torsion idxes.
        idx_batch = torch.arange(len(idx_i),device=device)
        idx_k_n = adj_t[idx_j].storage.col()
        repeat = num_triplets
        num_triplets_t = num_triplets.repeat_interleave(repeat)[mask]
        idx_i_t = idx_i.repeat_interleave(num_triplets_t)
        idx_j_t = idx_j.repeat_interleave(num_triplets_t)
        idx_k_t = idx_k.repeat_interleave(num_triplets_t)
        idx_batch_t = idx_batch.repeat_interleave(num_triplets_t)
        mask = idx_i_t != idx_k_n   
        idx_i_t, idx_j_t, idx_k_t, idx_k_n, idx_batch_t = idx_i_t[mask], idx_j_t[mask], idx_k_t[mask], idx_k_n[mask], idx_batch_t[mask]

        # Calculate torsions.
        pos_j0 = pos[idx_k_t] - pos[idx_j_t]
        pos_ji = pos[idx_i_t] - pos[idx_j_t]
        pos_jk = pos[idx_k_n] - pos[idx_j_t]
        dist_ji = pos_ji.pow(2).sum(dim=-1).sqrt()
        plane1 = torch.cross(pos_ji, pos_j0)
        plane2 = torch.cross(pos_ji, pos_jk)
        a = (plane1 * plane2).sum(dim=-1) # cos_angle * |plane1| * |plane2|
        b = (torch.cross(plane1, plane2) * pos_ji).sum(dim=-1) / dist_ji 
        torsion1 = torch.atan2(b, a) # -pi to pi
        torsion1[torsion1<=0]+=2*PI # 0 to 2pi
        torsion = scatter(torsion1,idx_batch_t,reduce='min')

        return dist, angle, torsion, i, j, idx_kj, idx_ji
    
    else:
        return dist, angle, i, j, idx_kj, idx_ji





class SphereNet(torch.nn.Module):
    r"""
         The spherical message passing neural network SphereNet from the `"Spherical Message Passing for 3D Molecular Graphs" <https://openreview.net/forum?id=givsRXsOt9r>`_ paper.
        
        Args:
            energy_and_force (bool, optional): If set to :obj:`True`, will predict energy and take the negative of the derivative of the energy with respect to the atomic positions as predicted forces. (default: :obj:`False`)
            cutoff (float, optional): Cutoff distance for interatomic interactions. (default: :obj:`5.0`)
            num_layers (int, optional): Number of building blocks. (default: :obj:`4`)
            hidden_channels (int, optional): Hidden embedding size. (default: :obj:`128`)
            out_channels (int, optional): Size of each output sample. (default: :obj:`1`)
            int_emb_size (int, optional): Embedding size used for interaction triplets. (default: :obj:`64`)
            basis_emb_size_dist (int, optional): Embedding size used in the basis transformation of distance. (default: :obj:`8`)
            basis_emb_size_angle (int, optional): Embedding size used in the basis transformation of angle. (default: :obj:`8`)
            basis_emb_size_torsion (int, optional): Embedding size used in the basis transformation of torsion. (default: :obj:`8`)
            out_emb_channels (int, optional): Embedding size used for atoms in the output block. (default: :obj:`256`)
            num_spherical (int, optional): Number of spherical harmonics. (default: :obj:`7`)
            num_radial (int, optional): Number of radial basis functions. (default: :obj:`6`)
            envelop_exponent (int, optional): Shape of the smooth cutoff. (default: :obj:`5`)
            num_before_skip (int, optional): Number of residual layers in the interaction blocks before the skip connection. (default: :obj:`1`)
            num_after_skip (int, optional): Number of residual layers in the interaction blocks before the skip connection. (default: :obj:`2`)
            num_output_layers (int, optional): Number of linear layers for the output blocks. (default: :obj:`3`)
            act: (function, optional): The activation funtion. (default: :obj:`swish`)
            output_init: (str, optional): The initialization fot the output. It could be :obj:`GlorotOrthogonal` and :obj:`zeros`. (default: :obj:`GlorotOrthogonal`)
            
    """
    def __init__(
        self, energy_and_force=False, cutoff=6.0, num_layers=4,
        hidden_channels=128, out_channels=1, int_emb_size=64,
        basis_emb_size_dist=8, basis_emb_size_angle=8, basis_emb_size_torsion=8, out_emb_channels=256,
        num_spherical=7, num_radial=6, envelope_exponent=5,
        num_before_skip=1, num_after_skip=2, num_output_layers=3,
        act=swish, output_init='GlorotOrthogonal', one_hot=True):
        super(SphereNet, self).__init__()

        self.cutoff = cutoff
        self.energy_and_force = energy_and_force
        self.one_hot = one_hot
        self.init_e = init(num_radial, hidden_channels, act, one_hot=one_hot)
        self.init_v = update_v(hidden_channels, out_emb_channels, out_channels, num_output_layers, act, output_init)
        self.init_u = update_u()
        self.emb = emb(num_spherical, num_radial, self.cutoff, envelope_exponent)

        self.update_vs = torch.nn.ModuleList([
            update_v(hidden_channels, out_emb_channels, out_channels, num_output_layers, act, output_init) for _ in range(num_layers)])

        self.update_es = torch.nn.ModuleList([
            update_e(hidden_channels, int_emb_size, basis_emb_size_dist, basis_emb_size_angle, basis_emb_size_torsion, num_spherical, num_radial, num_before_skip, num_after_skip,act) for _ in range(num_layers)])

        self.update_us = torch.nn.ModuleList([update_u() for _ in range(num_layers)])

        self.reset_parameters()

    def reset_parameters(self):
        self.init_e.reset_parameters()
        self.init_v.reset_parameters()
        self.emb.reset_parameters()
        for update_e in self.update_es:
            update_e.reset_parameters()
        for update_v in self.update_vs:
            update_v.reset_parameters()


    def forward(self, batch_data):
        x, z, pos, batch, edge_index = batch_data.x , batch_data.x_one_hot, batch_data.pos, batch_data.batch, batch_data.edge_index

        if self.energy_and_force:
            pos.requires_grad_()
            
        num_nodes=z.size(0)
        dist, angle, torsion, i, j, idx_kj, idx_ji = xyz_to_dat(pos, edge_index, num_nodes, use_torsion=True)

        emb = self.emb(dist, angle, torsion, idx_kj)

        #Initialize edge, node, graph features
        if self.one_hot:
            z = torch.argmax(z, dim=1)
            e = self.init_e(z, emb, i, j)
        else:
            e = self.init_e(x, emb, i, j)
            
        v = self.init_v(e, i)
        u = self.init_u(torch.zeros_like(scatter(v, batch, dim=0)), v, batch) #scatter(v, batch, dim=0)

        for update_e, update_v, update_u in zip(self.update_es, self.update_vs, self.update_us):
            e = update_e(e, emb, idx_kj, idx_ji)
            v = update_v(e, i)
            u = update_u(u, v, batch) #u += scatter(v, batch, dim=0)



        return u
    


class SphereNet_Model(torch.nn.Module):

    def __init__(
            self,
            energy_and_force,
            cutoff,
            num_layers,
            hidden_channels,
            out_channels,
            int_emb_size,
            basis_emb_size_dist,
            basis_emb_size_angle,
            basis_emb_size_torsion,
            out_emb_channels,
            num_spherical,
            num_radial,
            envelope_exponent,
            num_before_skip,
            num_after_skip,
            num_output_layers,
            one_hot = False
    ):
        super(SphereNet_Model, self).__init__()
        self.encoder = SphereNet(
            energy_and_force = energy_and_force,
            cutoff = cutoff,
            num_layers = num_layers,
            hidden_channels = hidden_channels,
            out_channels = out_channels,
            int_emb_size = int_emb_size,
            basis_emb_size_dist = basis_emb_size_dist,
            basis_emb_size_angle = basis_emb_size_angle,
            basis_emb_size_torsion = basis_emb_size_torsion,
            out_emb_channels = out_emb_channels,
            num_spherical = num_spherical,
            num_radial = num_radial,
            envelope_exponent = envelope_exponent,
            num_before_skip = num_before_skip,
            num_after_skip = num_after_skip,
            num_output_layers = num_output_layers,
            one_hot=one_hot
        )

        self.ffn_dropout = Dropout(p=0.25)
        self.lin1 = Linear(out_channels, 64)
        self.lin2 = Linear(64, 1)
        self.activate_func = ReLU()
        self.ff_dropout = Dropout(p=0.25)

    def forward(self, batch_data):
        
        # atom_index = torch.argmax(batch_data.one_hot_atom, dim = 1)
        graph_embedding = self.encoder(batch_data)

        graph_embedding = self.ffn_dropout(graph_embedding)
        prediction = self.lin2(self.activate_func(self.ff_dropout(self.lin1(graph_embedding))))
        
        return prediction


import torch
import torch.nn.functional as F
from torch.nn import Embedding, Sequential, Linear
from torch_scatter import scatter
from torch_geometric.nn import radius_graph
from math import pi as PI
from torch_geometric.data import Data   
from torch.nn import Linear, Sequential, BatchNorm1d, ReLU
from torch.nn import Dropout, Linear


# need to create a new embedding function for the 28-dim node features which contain discrete and continuous features
class FeatureEmbedding(torch.nn.Module):
    def __init__(self, input_dim, output_dim):
        super(FeatureEmbedding, self).__init__()
        self.network = Sequential(
            Linear(input_dim, output_dim // 2),
            BatchNorm1d(output_dim // 2),
            ReLU(),
            Linear(output_dim // 2, output_dim),
            ReLU()
        )
    
    def forward(self, x):
        return self.network(x)

    def reset_parameters(self):
        for module in self.network:
            if hasattr(module, 'reset_parameters'):
                module.reset_parameters()

class update_e(torch.nn.Module):
    def __init__(self, hidden_channels, num_filters, num_gaussians, cutoff):
        super(update_e, self).__init__()
        self.cutoff = cutoff
        self.lin = Linear(hidden_channels, num_filters, bias=False)
        self.mlp = Sequential(
            Linear(num_gaussians, num_filters),
            ShiftedSoftplus(),
            Linear(num_filters, num_filters),
        )
        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.lin.weight)
        torch.nn.init.xavier_uniform_(self.mlp[0].weight)
        self.mlp[0].bias.data.fill_(0)
        torch.nn.init.xavier_uniform_(self.mlp[2].weight)
        self.mlp[0].bias.data.fill_(0)

    def forward(self, v, dist, dist_emb, edge_index):
        j, _ = edge_index
        C = 0.5 * (torch.cos(dist * PI / self.cutoff) + 1.0)
        W = self.mlp(dist_emb) * C.view(-1, 1)
        v = self.lin(v)
        e = v[j] * W
        return e

class update_v(torch.nn.Module):
    def __init__(self, hidden_channels, num_filters):
        super(update_v, self).__init__()
        self.act = ShiftedSoftplus()
        self.lin1 = Linear(num_filters, hidden_channels)
        self.lin2 = Linear(hidden_channels, hidden_channels)

        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.lin1.weight)
        self.lin1.bias.data.fill_(0)
        torch.nn.init.xavier_uniform_(self.lin2.weight)
        self.lin2.bias.data.fill_(0)

    def forward(self, v, e, edge_index):
        _, i = edge_index
        out = scatter(e, i, dim=0)
        out = self.lin1(out)
        out = self.act(out)
        out = self.lin2(out)
        return v + out

class update_u(torch.nn.Module):
    def __init__(self, hidden_channels, out_channels):
        super(update_u, self).__init__()
        self.lin1 = Linear(hidden_channels, hidden_channels // 2)
        self.act = ShiftedSoftplus()
        self.lin2 = Linear(hidden_channels // 2, out_channels)

        self.reset_parameters()

    def reset_parameters(self):
        torch.nn.init.xavier_uniform_(self.lin1.weight)
        self.lin1.bias.data.fill_(0)
        torch.nn.init.xavier_uniform_(self.lin2.weight)
        self.lin2.bias.data.fill_(0)

    def forward(self, v, batch):
        v = self.lin1(v)
        v = self.act(v)
        v = self.lin2(v)
        u = scatter(v, batch, dim=0)
        return u

class emb(torch.nn.Module):
    def __init__(self, start=0.0, stop=5.0, num_gaussians=50):
        super(emb, self).__init__()
        offset = torch.linspace(start, stop, num_gaussians)
        self.coeff = -0.5 / (offset[1] - offset[0]).item() ** 2
        self.register_buffer('offset', offset)

    def forward(self, dist):
        dist = dist.view(-1, 1) - self.offset.view(1, -1)
        return torch.exp(self.coeff * torch.pow(dist, 2))

class ShiftedSoftplus(torch.nn.Module):
    def __init__(self):
        super(ShiftedSoftplus, self).__init__()
        self.shift = torch.log(torch.tensor(2.0)).item()

    def forward(self, x):
        return F.softplus(x) - self.shift

class SchnetLayer(torch.nn.Module):
    def __init__(self, energy_and_force=False, cutoff=6.0, num_layers=6, input_dim=28, hidden_channels=128, num_filters=128,
                 num_gaussians=50, out_channels=32, one_hot=False):
        # keep cutoff as 6.0 across all experiments
        
        super(SchnetLayer, self).__init__()
        self.energy_and_force = energy_and_force
        self.cutoff = cutoff
        self.num_layers = num_layers
        self.hidden_channels = hidden_channels
        self.num_filters = num_filters
        self.num_gaussians = num_gaussians
        self.one_hot = one_hot
        
        # Embedding for mixed feature types
        self.feature_embedding = FeatureEmbedding(input_dim, hidden_channels) 
        self.atom_embedding = Embedding(100, hidden_channels)
        
        self.dist_emb = emb(0.0, cutoff, num_gaussians)
        self.update_vs = torch.nn.ModuleList([update_v(hidden_channels, num_filters) for _ in range(num_layers)])
        self.update_es = torch.nn.ModuleList([update_e(hidden_channels, num_filters, num_gaussians, cutoff) for _ in range(num_layers)])
        self.update_u = update_u(hidden_channels, out_channels)
        self.reset_parameters()

    def reset_parameters(self):
        self.feature_embedding.network.apply(self.init_weights)
        self.atom_embedding.reset_parameters()
        for update_e in self.update_es:
            update_e.reset_parameters()
        for update_v in self.update_vs:
            update_v.reset_parameters()
        self.update_u.reset_parameters()

    @staticmethod
    def init_weights(m):
        if isinstance(m, torch.nn.Linear):
            torch.nn.init.xavier_uniform_(m.weight)
            torch.nn.init.constant_(m.bias, 0)

    def forward(self, batch_data):

        x, z, pos, batch, edge_index = batch_data.x , batch_data.x_one_hot, batch_data.pos, batch_data.batch, batch_data.edge_index
        if self.energy_and_force:
            pos.requires_grad_()

        row, col = edge_index
        dist = (pos[row] - pos[col]).norm(dim=-1)
        
        dist_emb = self.dist_emb(dist)
        if self.one_hot:
            z_index = torch.argmax(z, dim=1)
            v = self.atom_embedding(z_index)
        else:
            v = self.feature_embedding(x)
            
        for update_e, update_v in zip(self.update_es, self.update_vs):
            e = update_e(v, dist, dist_emb, edge_index)
            v = update_v(v, e, edge_index)

        u = self.update_u(v, batch)
        
        return u

### Schnet Model

class SchNet_Model(torch.nn.Module):
    def __init__(
            self, energy_and_force=False, cutoff=6.0, num_layers=6, in_channels = 28, hidden_channels=128, num_filters=128,
            num_gaussians=50, out_channels=32, one_hot=False):

        super(SchNet_Model, self).__init__()
        self.encoder = SchnetLayer(energy_and_force=energy_and_force,
                        cutoff=cutoff,
                        input_dim=in_channels,
                        num_layers=num_layers,
                        hidden_channels=hidden_channels,
                        num_filters=num_filters,
                        num_gaussians=num_gaussians,
                        out_channels=out_channels,
                        one_hot=one_hot
                    )
        self.one_hot = one_hot
        self.ffn_dropout = Dropout(p=0.25)
        self.lin1 = Linear(out_channels, 64)
        self.lin2 = Linear(64, 1)
        self.activate_func = ReLU()
        self.ff_dropout = Dropout(p=0.25)

    def forward(self, batch_data):
        graph_embedding = self.encoder(batch_data)

        graph_embedding = self.ffn_dropout(graph_embedding)
        prediction = self.lin2(self.activate_func(self.ff_dropout(self.lin1(graph_embedding))))
        
        return prediction





import torch
import random
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

class Smiles2Vec(nn.Module):
    def __init__(self,
                 char_to_idx,
                 max_seq_len=270,
                 embedding_dim=50,
                 use_bidir=True,
                 use_conv=True,
                 filters=192,
                 kernel_size=3,
                 strides=1,
                 rnn_sizes=[300, 300,300],
                 rnn_types=["GRU", "GRU", "GRU"],
                 ):
        super(Smiles2Vec, self).__init__()
        
        self.char_to_idx = char_to_idx
        self.max_seq_len = max_seq_len
        self.embedding_dim = embedding_dim
        self.use_bidir = use_bidir
        self.use_conv = use_conv
        self.rnn_sizes = rnn_sizes
        self.rnn_types = rnn_types
        # self.mode = mode

        self.embedding = nn.Embedding(num_embeddings=len(char_to_idx), embedding_dim=embedding_dim)

        if use_conv:
            self.conv1d = nn.Conv1d(in_channels=embedding_dim,
                                    out_channels=filters,
                                    kernel_size=kernel_size,
                                    stride=strides)
        
        self.rnn_layers = nn.ModuleList()
        input_size = filters if use_conv else embedding_dim
        for idx, rnn_type in enumerate(rnn_types):
            rnn_class = nn.GRU if rnn_type == "GRU" else nn.LSTM
            rnn_layer = rnn_class(input_size, rnn_sizes[idx], batch_first=True, bidirectional=use_bidir)
            self.rnn_layers.append(rnn_layer)
            input_size = rnn_sizes[idx] * 2 if use_bidir else rnn_sizes[idx]

        self.fc = nn.Linear(input_size, 1)
        # if mode == "classification":
        #     self.sigmoid = nn.Sigmoid()
    
    def forward(self, batch_data):
        smiles_list = batch_data.smiles
        smiles_seqs = []
        for smiles in smiles_list:
            seq = [self.char_to_idx[char] for char in smiles]
            seq += [0] * (self.max_seq_len - len(seq))
            smiles_seqs.append(seq)
        
        x = torch.tensor(smiles_seqs, dtype=torch.long, device=batch_data.x.device)

        x = self.embedding(x)
        
        if self.use_conv:
            x = x.permute(0, 2, 1)  # Conv1d expects (batch, channels, seq_len)
            x = F.relu(self.conv1d(x))
            x = x.permute(0, 2, 1)  # Back to (batch, seq_len, channels)
        
        for rnn_layer in self.rnn_layers:
            x, _ = rnn_layer(x)
        
        output = self.fc(x[:, -1, :])
        # if self.mode == "classification":
        #     output = self.sigmoid(output)
        return output

# Test case
if __name__ == "__main__":
    # Sample character-to-index mapping
    # Set random seeds for reproducibility
    seed = 42
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    
    char_to_idx = {'[': 0, 'P': 1, 'e': 2, '@': 3, '1': 4, '#': 5, 
                   'n': 6, 'S': 7, '+': 8, '=': 9, ']': 10, 'c': 11, 
                   'o': 12, '2': 13, 'B': 14, '\\': 15, 'O': 16, 'H': 17, 
                   '5': 18, '-': 19, 'l': 20, '6': 21, 's': 22, '/': 23, 'C': 24, 
                   '4': 25, '3': 26, 'N': 27, 'I': 28, 'F': 29, '(': 30, ')': 31, 'r': 32}

    # Sample SMILES strings
    smiles_strings = ["Cc1ccc(C)c(OCC(O)CN2C(C)CCCC2C)c1", "CCOC(=O)N1CCN(C2=Nc3cccc4cccc2c34)CC1"]

    # Convert SMILES strings to integer sequences
    max_seq_len = 37  # Define the maximum sequence length
    smiles_seqs = []
    for smiles in smiles_strings:
        seq = [char_to_idx[char] for char in smiles]
        # Pad sequences to the maximum sequence length
        seq += [0] * (max_seq_len - len(seq))
        smiles_seqs.append(seq)

    smiles_seqs = torch.tensor(smiles_seqs, dtype=torch.long)

    # Instantiate the model
    model = Smiles2Vec(char_to_idx, max_seq_len=max_seq_len, mode="classification")

    # Pass the batch of SMILES sequences through the model
    outputs = model(smiles_seqs)

    # Print the outputs
    print("Model outputs:")
    print(outputs)



import math

allowable_features = {
    'possible_atomic_num_list' : list(range(1, 119)) + ['misc'],
    'possible_chirality_list' : [
        'CHI_UNSPECIFIED',
        'CHI_TETRAHEDRAL_CW',
        'CHI_TETRAHEDRAL_CCW',
        'CHI_OTHER',
        'misc'
    ],
    'possible_degree_list' : [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 'misc'],
    'possible_formal_charge_list' : [-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 'misc'],
    'possible_numH_list' : [0, 1, 2, 3, 4, 5, 6, 7, 8, 'misc'],
    'possible_number_radical_e_list': [0, 1, 2, 3, 4, 'misc'],
    'possible_hybridization_list' : [
        'SP', 'SP2', 'SP3', 'SP3D', 'SP3D2', 'misc'
        ],
    'possible_is_aromatic_list': [False, True],
    'possible_is_in_ring_list': [False, True],
    'possible_bond_type_list' : [
        'SINGLE',
        'DOUBLE',
        'TRIPLE',
        'AROMATIC',
        'misc'
    ],
    'possible_bond_stereo_list': [
        'STEREONONE',
        'STEREOZ',
        'STEREOE',
        'STEREOCIS',
        'STEREOTRANS',
        'STEREOANY',
    ], 
    'possible_is_conjugated_list': [False, True],
}

def one_hot_vector(val, lst):
	'''
	Converts a value to a one-hot vector based on options in lst
	'''
	if val not in lst:
		val = lst[-1]
	return map(lambda x: x == val, lst)

def safe_index(l, e):
    """
    Return index of element e in list l. If e is not present, return the last index
    """
    try:
        return l.index(e)
    except:
        return len(l) - 1


def atom_to_feature_vector(atom):
    """
    Converts rdkit atom object to feature list of indices
    :param mol: rdkit atom object
    :return: list
    """
    atom_feature = []
    atom_feature += one_hot_vector(atom.GetAtomicNum(), [1, 6, 7, 8, 9, 14, 15, 16, 17, 35, 53, 999])
    atom_feature += one_hot_vector(len(atom.GetNeighbors()), list(range(1, 5)))
    atom_feature.append(atom.GetFormalCharge())
    atom_feature.append(atom.IsInRing())
    atom_feature.append(atom.GetIsAromatic())
    atom_feature.append(atom.GetExplicitValence())
    atom_feature.append(atom.GetMass())

    # Add Gasteiger charge and set to 0 if it is NaN or Infinite
    gasteiger_charge = float(atom.GetProp('_GasteigerCharge'))
    if math.isnan(gasteiger_charge) or math.isinf(gasteiger_charge):
        gasteiger_charge = 0
    atom_feature.append(gasteiger_charge)

    # Add Gasteiger H charge and set to 0 if it is NaN or Infinite
    gasteiger_h_charge = float(atom.GetProp('_GasteigerHCharge'))
    if math.isnan(gasteiger_h_charge) or math.isinf(gasteiger_h_charge):
        gasteiger_h_charge = 0

    atom_feature.append(gasteiger_h_charge)

    return atom_feature

def atom_to_one_hot_vector(atom):
    atom_feature = []
    atom_feature += one_hot_vector(atom.GetAtomicNum(), [1, 6, 7, 8, 9, 14, 15, 16, 17, 35, 53, 999])

    return atom_feature



def get_atom_feature_dims():
    return list(map(len, [
        allowable_features['possible_atomic_num_list'],
        allowable_features['possible_chirality_list'],
        allowable_features['possible_degree_list'],
        allowable_features['possible_formal_charge_list'],
        allowable_features['possible_numH_list'],
        allowable_features['possible_number_radical_e_list'],
        allowable_features['possible_hybridization_list'],
        allowable_features['possible_is_aromatic_list'],
        allowable_features['possible_is_in_ring_list']
        ]))

def bond_to_feature_vector(bond):
    """
    Converts rdkit bond object to feature list of indices
    :param mol: rdkit bond object
    :return: list
    """
    bond_feature = [
                safe_index(allowable_features['possible_bond_type_list'], str(bond.GetBondType())),
                allowable_features['possible_bond_stereo_list'].index(str(bond.GetStereo())),
                allowable_features['possible_is_conjugated_list'].index(bond.GetIsConjugated()),
            ]
    return bond_feature


def get_bond_feature_dims():
    return list(map(len, [
        allowable_features['possible_bond_type_list'],
        allowable_features['possible_bond_stereo_list'],
        allowable_features['possible_is_conjugated_list']
        ]))

def atom_feature_vector_to_dict(atom_feature):
    [atomic_num_idx, 
    chirality_idx,
    degree_idx,
    formal_charge_idx,
    num_h_idx,
    number_radical_e_idx,
    hybridization_idx,
    is_aromatic_idx,
    is_in_ring_idx] = atom_feature

    feature_dict = {
        'atomic_num': allowable_features['possible_atomic_num_list'][atomic_num_idx],
        'chirality': allowable_features['possible_chirality_list'][chirality_idx],
        'degree': allowable_features['possible_degree_list'][degree_idx],
        'formal_charge': allowable_features['possible_formal_charge_list'][formal_charge_idx],
        'num_h': allowable_features['possible_numH_list'][num_h_idx],
        'num_rad_e': allowable_features['possible_number_radical_e_list'][number_radical_e_idx],
        'hybridization': allowable_features['possible_hybridization_list'][hybridization_idx],
        'is_aromatic': allowable_features['possible_is_aromatic_list'][is_aromatic_idx],
        'is_in_ring': allowable_features['possible_is_in_ring_list'][is_in_ring_idx]
    }

    return feature_dict


def bond_feature_vector_to_dict(bond_feature):
    [bond_type_idx, 
    bond_stereo_idx,
    is_conjugated_idx] = bond_feature

    feature_dict = {
        'bond_type': allowable_features['possible_bond_type_list'][bond_type_idx],
        'bond_stereo': allowable_features['possible_bond_stereo_list'][bond_stereo_idx],
        'is_conjugated': allowable_features['possible_is_conjugated_list'][is_conjugated_idx]
    }

    return feature_dict



import numpy as np
from sklearn.metrics import confusion_matrix, auc, roc_curve, f1_score, \
    roc_auc_score


def sigmoid(x):
    z = 1 / (1 + np.exp(-x))
    return z

def MAE(true_y, predicted_score):
    return np.mean(np.abs(true_y - predicted_score))

def MSE(true_y, predicted_score):
    return np.mean((true_y - predicted_score) ** 2)

def RMSE(true_y, predicted_score):
    return np.sqrt(np.mean((true_y - predicted_score) ** 2))

def R2(true_y, predicted_score):
    SS_res = np.sum((true_y - predicted_score) ** 2)
    SS_tot = np.sum((true_y - np.mean(true_y)) ** 2)
    return 1 - SS_res / SS_tot



def calculate_logAUC(true_y, predicted_score, FPR_range=(0.001, 0.1)):
    """
    Author: Yunchao "Lance" Liu (lanceknight26@gmail.com)
    Calculate logAUC in a certain FPR range (default range: [0.001, 0.1]).
    This was used by previous methods [1] and the reason is that only a
    small percentage of samples can be selected for experimental tests in
    consideration of cost. This means only molecules with very high
    predicted activity values can be worth testing, i.e., the decision
    threshold is high. And the high decision threshold corresponds to the
    left side of the ROC curve, i.e., those FPRs with small values. Also,
    because the threshold cannot be predetermined, the area under the curve
    is used to consolidate all possible thresholds within a certain FPR
    range. Finally, the logarithm is used to bias smaller FPRs. The higher
    the logAUC[0.001, 0.1], the better the performance.

    A perfect classifer gets a logAUC[0.001, 0.1] ) of 1, while a random
    classifer gets a logAUC[0.001, 0.1] ) of around 0.0215 (See [2])

    References:
    [1] Mysinger, M.M. and B.K. Shoichet, Rapid Context-Dependent Ligand
    Desolvation in Molecular Docking. Journal of Chemical Information and
    Modeling, 2010. 50(9): p. 1561-1573.
    [2] Mendenhall, J. and J. Meiler, Improving quantitative
    structure–activity relationship models using Artificial Neural Networks
    trained with dropout. Journal of computer-aided molecular design,
    2016. 30(2): p. 177-189.
    :param true_y: numpy array of the ground truth. Values are either 0 (
    inactive) or 1(active).
    :param predicted_score: numpy array of the predicted score (The
    score does not have to be between 0 and 1)
    :param FPR_range: the range for calculating the logAUC formated in
    (x, y) with x being the lower bound and y being the upper bound
    :return: a numpy array of logAUC of size [1,1]
    """

    np.seterr(divide='ignore')
    # FPR range validity check
    if FPR_range == None:
        raise Exception('FPR range cannot be None')
    lower_bound = FPR_range[0]
    upper_bound = FPR_range[1]
    if (lower_bound >= upper_bound):
        raise Exception('FPR upper_bound must be greater than lower_bound')

    fpr, tpr, thresholds = roc_curve(true_y, predicted_score, pos_label=1)

    tpr = np.append(tpr, np.interp([lower_bound, upper_bound], fpr, tpr))
    fpr = np.append(fpr, [lower_bound, upper_bound])

    # Sort both x-, y-coordinates array
    tpr = np.sort(tpr)
    fpr = np.sort(fpr)

    # Get the data points' coordinates. log_fpr is the x coordinate, tpr is the y coordinate.
    log_fpr = np.log10(fpr)
    x = log_fpr
    y = tpr
    lower_bound = np.log10(lower_bound)
    upper_bound = np.log10(upper_bound)

    # Get the index of the lower and upper bounds
    lower_bound_idx = np.where(x == lower_bound)[-1][-1]
    upper_bound_idx = np.where(x == upper_bound)[-1][-1]

    # Create a new array trimmed at the lower and upper bound
    trim_x = x[lower_bound_idx:upper_bound_idx + 1]
    trim_y = y[lower_bound_idx:upper_bound_idx + 1]

    area = auc(trim_x, trim_y) / (upper_bound - lower_bound)
    return area



def cal_EF(true_y, predicted_score, k):
    """
    Calculate the enrichment factor (EF) for a given predicted score and true labels.

    The enrichment factor is a measure of the enrichment of true positives in the top k predictions.
    It is defined as the ratio of the number of true positives in the top k predictions to the
    total number of true positives in the dataset, scaled by the ratio of the number of compounds
    in the selection set to the total number of compounds in the dataset.

    Args:
        true_y (array_like):
            Binary class labels. 1 for positive class, 0 otherwise.
        predicted_score (array_like):
            Prediction values.
    
    Returns:
        float:
            Enrichment factor value.
    """

    # Ns: the number of compounds in the selection set (the predicted actives), k
    # N: the number of compounds in the entire dataset
    # ns: the number of true active compounds in the selection set
    # n: the number of true active compounds in the entire dataset

    order_idx_pred = np.argsort(-predicted_score)

    order_predicted_score = predicted_score[order_idx_pred]
    order_true_y = true_y[order_idx_pred]

    N = len(true_y)
    Ns = k

    ns = (order_true_y[:k] == 1).sum()
    n = true_y.sum()

    return N * ns / (Ns * n)



def cal_DCG(true_y, predicted_score, k):
    # assuming k<= len(true_y) == len(predicted_score)
    """
    Calculate the discounted cumulative gain (DCG) for a given predicted score and true labels.

    The DCG is a measure of the cumulative gain of true positives in the top k predictions.
    It is defined as the sum of the true positives in the top k predictions, discounted by the
    logarithm of their position in the ranking.

    Args:
        true_y (array_like):
            Binary class labels. 1 for positive class, 0 otherwise.
        predicted_score (array_like):
            Prediction values.
        k (int):
            The number of top predictions to consider.
    
    Returns:
        float:
            Discounted cumulative gain value.
    """

    order_idx_pred = np.argsort(-predicted_score)
    order_idx_label = np.argsort(-true_y)

    dcg = np.sum(true_y[order_idx_pred][:k] / np.log2(np.arange(1 + 1, (k + 1) + 1)))

    return dcg


def cal_BEDROC_score(true_y, predicted_score, decreasing=True, alpha=20.0):
    """BEDROC metric implemented according to Truchon and Bayley.

    The Boltzmann Enhanced Descrimination of the Receiver Operator
    Characteristic (BEDROC) score is a modification of the Receiver Operator
    Characteristic (ROC) score that allows for a factor of *early recognition*.

    References:
        The original paper by Truchon et al. is located at `10.1021/ci600426e
        <http://dx.doi.org/10.1021/ci600426e>`_.

    Args:
        true_y (array_like):
            Binary class labels. 1 for positive class, 0 otherwise.
        predicted_score (array_like):
            Prediction values.
        decreasing (bool):
            True if high values of ``predicted_score`` correlates to positive class.
        alpha (float):
            Early recognition parameter.

    Returns:
        float:
            Value in interval [0, 1] indicating degree to which the predictive
            technique employed detects (early) the positive class.
     """

    assert len(true_y) == len(predicted_score), \
        'The number of scores must be equal to the number of labels'

    big_n = len(true_y)
    n = sum(true_y == 1)

    if decreasing:
        order = np.argsort(-predicted_score)
    else:
        order = np.argsort(predicted_score)

    m_rank = (true_y[order] == 1).nonzero()[0]

    s = np.sum(np.exp(-alpha * m_rank / big_n))

    r_a = n / big_n

    rand_sum = r_a * (1 - np.exp(-alpha)) / (np.exp(alpha / big_n) - 1)

    fac = r_a * np.sinh(alpha / 2) / (np.cosh(alpha / 2) -
                                      np.cosh(alpha / 2 - alpha * r_a))

    cte = 1 / (1 - np.exp(alpha * (1 - r_a)))

    return s * fac / rand_sum + cte


import os
import numpy as np
from tqdm import tqdm
from welqrate.utils.evaluation import calculate_logAUC, cal_EF, cal_DCG, cal_BEDROC_score

def get_test_metrics(model, loader, device, type = 'test', 
                     save_per_molecule_pred=False, save_path=None, extra_metrics=False):
    model.eval()

    all_pred_y = []
    all_true_y = []

    for i, batch in enumerate(tqdm(loader)):
        batch.to(device)
        pred_y = model(batch).cpu().view(-1).detach().numpy()
        true_y = batch.y.view(-1).cpu().numpy()
        
        for j, _ in enumerate(pred_y):
            all_pred_y.append(pred_y[j])
            all_true_y.append(true_y[j])
    
    if save_per_molecule_pred and save_path is not None:
        filename = os.path.join(save_path, f'per_molecule_pred_of_{type}_set.txt')
        with open(filename, 'w') as out_file:
            for k, _ in enumerate(all_pred_y):
                out_file.write(f'{all_pred_y[k]}\ty={all_true_y[k]}\n')
        
        # rank prediction
        with open(filename, 'r') as f:
            data = [(float(line.split('\t')[0]), line.split('\t')[1] ) for line in f.readlines()]
        ranked_data = sorted(data, key=lambda x: x[0], reverse=True)
        with open(os.path.join(save_path, f'ranked_mol_score_{type}.txt'), 'w') as f:
            for i, (score, label) in enumerate(ranked_data):
                f.write(f"{i}\t{score}\t{label}")

    all_pred_y = np.array(all_pred_y)
    all_true_y = np.array(all_true_y)
    logAUC = calculate_logAUC(all_true_y, all_pred_y)
    EF100 = cal_EF(all_true_y, all_pred_y, 100)
    DCG100 = cal_DCG(all_true_y, all_pred_y, 100)
    BEDROC = cal_BEDROC_score(all_true_y, all_pred_y)

    if extra_metrics:
        EF500 = cal_EF(all_true_y, all_pred_y, 500)
        EF1000 = cal_EF(all_true_y, all_pred_y, 1000)
        DCG500 = cal_DCG(all_true_y, all_pred_y, 500)
        DCG1000 = cal_DCG(all_true_y, all_pred_y, 1000)
        return logAUC, EF100, DCG100, BEDROC, EF500, EF1000, DCG500, DCG1000
    else:
        return logAUC, EF100, DCG100, BEDROC


from torch.nn import Module, Linear, Sequential, Sigmoid, Dropout, ReLU
import torch
from torch_scatter import scatter_mean, scatter_add


from torch.nn import Module, Linear, ReLU, Dropout, Sequential

class MLP(Module):
    def __init__(self, input_dim, hidden_dim=32, num_layers=2, one_hot=False):
        super().__init__()
        self.one_hot = one_hot

        # Setup the MLP layers using ReLU activation
        layers = []
        for _ in range(num_layers):
            layers.extend([
                Dropout(0.05),
                Linear(input_dim, hidden_dim),
                ReLU(),
                Dropout(0.35),
            ])
            input_dim = hidden_dim  

        layers.append(Linear(hidden_dim, 1))  # Final output layer

        self.mlp = Sequential(*layers)


    def forward(self, batch_data):
        if self.one_hot:
            x = batch_data.one_hot_atom.float()  
        else:
            x = batch_data.x
          
        x = scatter_add(x, batch_data.batch, dim=0)
        x = self.mlp(x)

        return x

class bcl_MLP(Module):
    def __init__(self, hidden_dim=32, num_layers=2, input_dim=391):
        super().__init__()

        # Setup the MLP layers using ReLU activation
        layers = []
        for _ in range(num_layers):
            layers.extend([
                Dropout(0.05),
                Linear(input_dim, hidden_dim),
                ReLU(),
                Dropout(0.35),
            ])
            input_dim = hidden_dim  # Update input dimension for the next layer

        layers.append(Linear(hidden_dim, 1))  # Final output layer

        self.mlp = Sequential(*layers)

    def forward(self, batch_data):
        x = batch_data.bcl.view(-1, 391)
        x = self.mlp(x)

        return x






# Adapted from the scheduler used in Graphormer:
# https://github.com/microsoft/Graphormer/tree/ogb-lsc/OGB-LSC/graphormer/src
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import math
from torch.optim.lr_scheduler import _LRScheduler

class PolynomialDecayLR(_LRScheduler):

    def __init__(self, optimizer, warmup_iterations, tot_iterations, lr, end_lr, power, last_epoch=-1, verbose=False):
        self.warmup_updates = warmup_iterations
        self.tot_updates = tot_iterations
        self.lr = lr
        self.end_lr = end_lr
        self.power = power
        super(PolynomialDecayLR, self).__init__(optimizer, last_epoch, verbose)

    def get_lr(self):
        if self._step_count <= self.warmup_updates:
            self.warmup_factor = self._step_count / float(self.warmup_updates)
            lr = self.warmup_factor * self.lr
        elif self._step_count >= self.tot_updates:
            lr = self.end_lr
        else:
            warmup = self.warmup_updates
            lr_range = self.lr - self.end_lr
            pct_remaining = 1 - (self._step_count - warmup) / (
                self.tot_updates - warmup
            )
            lr = lr_range * pct_remaining ** (self.power) + self.end_lr

        return [lr for group in self.optimizer.param_groups]

    def _get_closed_form_lr(self):
        assert False

def get_scheduler(optimizer, config, train_loader):
    # Calculate total iterations correctly
    num_train_batches = len(train_loader)
    tot_iterations = num_train_batches * int(config['train']['num_epochs'])
    print(f'tot_iterations={tot_iterations}')
    
    # Calculate warmup iterations as a percentage of total iterations
    warmup_percentage = float(config['train']['warmup_percentage'])  
    warmup_iterations = max(1, int(warmup_percentage * tot_iterations))
    print(f'warmup_iterations={warmup_iterations}')
    
    scheduler = {
        'scheduler': PolynomialDecayLR(
            optimizer,
            warmup_iterations=warmup_iterations,
            tot_iterations=tot_iterations,
            lr=float(config['train']['peak_lr']),
            end_lr=float(config['train']['end_lr']),
            power=1.0,
        ),
        'interval': 'step',
        'frequency': 1,
    }
    return scheduler['scheduler']

def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group['lr']


import matplotlib.pyplot as plt
import os

def plot_epoch():
    # Define the filename and load the data
    filename = "logs/loss_per_epoch.log"
    with open(filename, "r") as f:
        lines = f.readlines()

    epochs = []
    losses = []
    logAUCs = []
    lrs =[]
    # ts = []

    fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(12, 5))
    # Parse the data from each line
    for line in lines:

        # epoch, loss,  lr, t = line.strip().split("\t")
        if len(line.strip().split("\t"))==3:
            epoch, loss, lr = line.strip().split("\t")
        elif len(line.strip().split("\t"))==4:
            epoch, loss, logAUC, lr = line.strip().split("\t")
            logAUCs.append(float(logAUC.split("=")[1]))
        epochs.append(int(epoch))
        losses.append(float(loss.split("=")[1]))
        lrs.append(float(lr.split("=")[1]))
        # ts.append(float(t.split("=")[1]))

    # Plot the loss vs epoch
    ax1.scatter(epochs, losses)
    ax1.set_xlabel("Epoch")
    ax1.set_ylabel("Loss")

    # Plot the logAUC vs epoch
    if len(line.strip().split("\t")) == 4:
        print(f'epoch={epoch}, logAUC={logAUCs}')
        ax2.scatter(epochs, logAUCs)
    ax2.set_xlabel("Epoch")
    ax2.set_ylabel("logAUC")

    ax3.scatter(epochs, lrs)
    ax3.set_xlabel("Epoch")
    ax3.set_ylabel("learning rate")

    # ax4.scatter(epochs, ts)
    # ax4.set_xlabel("Epoch")
    # ax4.set_ylabel("temperature")

    # Save the plot as a PNG file
    os.makedirs('images', exist_ok=True)
    plt.savefig("images/loss_and_auc_vs_epoch.png")

if __name__ == "__main__":
    plot_epoch()






def gcn_objective(trial):
    # Define hyperparameter search space
    hidden_channels = trial.suggest_categorical('hidden_channels', [32, 64, 128])
    num_layers = trial.suggest_int('num_layers', 2, 4)
    peak_lr = trial.suggest_float('peak_lr', 1e-4, 1e-2, log=True)
    trial_dir = f"{results_dir}/{dataset_name}/{split_scheme}/gcn/trial{trial.number}"
    os.makedirs(trial_dir, exist_ok=True)

    try:
        # Update config
        config = copy.deepcopy(base_config)
        config['model']['hidden_channels'] = hidden_channels
        config['model']['num_layers'] = num_layers
        config['train']['peak_lr'] = peak_lr
        config['data']['split_scheme'] = args.split
        config['train']['num_epochs'] = args.num_epochs
        
        # Initialize model with current params
        model = GCN_Model(
            in_channels=12,
            hidden_channels=hidden_channels,
            num_layers=num_layers,
        ).to(device)
        
        print(f"\nTrial {trial.number}")
        print(f"Hidden channels: {hidden_channels}")
        print(f"Number of layers: {num_layers}")
        print(f"Peak learning rate: {peak_lr}")
        
        # Train model and get metrics
        test_logAUC, test_EF100, test_DCG100, test_BEDROC, test_EF500, test_EF1000, test_DCG500, test_DCG1000 = train(
            model=model, 
            train_loader=train_loader,
            valid_loader=valid_loader,
            test_loader=test_loader,
            config=config, 
            device=device,
            save_path=trial_dir,
        )
        
        # Save results to CSV
        with open(csv_file, 'a', newline='') as f:
            writer = csv.writer(f)
            writer.writerow([hidden_channels, num_layers, peak_lr, 
                            test_logAUC, test_EF100, test_DCG100, test_BEDROC,
                            test_EF500, test_EF1000, test_DCG500, test_DCG1000])
        
        return test_BEDROC
        
    except Exception as e:
        print(f"Error occurred in trial {trial.number}")
        print(f"Error message: {str(e)}")
        # Save error info to CSV
        with open(csv_file, 'a', newline='') as f:
            writer = csv.writer(f)
            writer.writerow([hidden_channels, num_layers, peak_lr, 
                            float('-inf'), float('-inf'), float('-inf'), float('-inf'),
                            float('-inf'), float('-inf'), float('-inf'), float('-inf')])
        return float('-inf')  


from torch_geometric.data import InMemoryDataset
from torch_geometric.data import Dataset, Data
import torch
import os
import os.path as osp
from tqdm import tqdm
from rdkit import Chem, RDLogger
import pandas as pd
from welqrate.mol_utils.preprocess import (smiles2graph, inchi2graph, sdffile2mol_conformer,
                                  mol_conformer2graph3d)
from glob import glob
import subprocess


class WelQrateDataset(InMemoryDataset):
    r"""
    Nine WelQrate Datasets from the Paper: 
    `"WelQrate: Defining the Gold Standard in Small Molecule Drug Discovery" 
    https://arxiv.org/pdf/2411.09820`
    
    Args:
        dataset_name (str): Name of the WelQrate dataset to load, e.g., 'AID1798', 'AID1843'
        mol_repr (str): Molecule representation, either '2d_graph' or '3d_graph'
        source (str): Source of the raw molecule data for 2D graph, either 'smiles' or 'inchi'

    Returns:
        Common attributes:
            y: label of the molecule
            smiles: smiles of the molecule
            inchi: inchi of the molecule
            pubchem_cid: pubchem cid of the molecule
            mol_id: id of the molecule in the dataset
        2D graph:
            x: atom features
            x_one_hot: one-hot encoding of the atom types
            edge_index: connectivity of the graph
            edge_attr: bond features
        3D graph:
            x: atom features
            x_one_hot: one-hot encoding of the atom types
            edge_index: connectivity of the graph calculated by radius graph
            pos: 3D coordinates of the atoms
            atom_num: atomic number of the atoms e.g., [1, 6, 8, 1, 1]
    """

    def __init__(self, 
                 dataset_name: str, 
                 root: str, 
                 mol_repr ='2d_graph', 
                 source ='smiles'):
        
        self.name = dataset_name
        self.root = osp.join(root, dataset_name)
        self.mol_repr = mol_repr
        self.source = source
        if source not in ['smiles', 'inchi']:
            raise ValueError(f'source must be either "smiles" or "inchi"')
        if mol_repr not in ['2d_graph', '3d_graph']:
            raise ValueError(f'mol_repr must be either "2d_graph" or "3d_graph"')
        
        print(f'dataset stored in {self.root}')
        super(WelQrateDataset, self).__init__(self.root)
        self.data, self.slices = torch.load(self.processed_paths[0])
        y = self.y.squeeze()
        self.num_active = len([x for x in y if x == 1])
        self.num_inactive = len([x for x in y if x == 0])

        print(f"Dataset {self.name} loaded.")
        print(f"Number of active molecules: {self.num_active}")
        print(f"Number of inactive molecules: {self.num_inactive}")

    @property
    def welqrate_dataset_catalog(self):
        return {
                "AID1798": {
                    "raw_url": "https://vanderbilt.box.com/shared/static/cd2dpdinu8grvi8dye3gi9bzsdt659d1.zip",
                    "file_type": "zip",
                    "raw_files": ['AID1798_actives.csv', 'AID1798_inactives.csv', 'AID1798_actives.sdf', 'AID1798_inactives.sdf'],
                    "split_url": "https://vanderbilt.box.com/shared/static/68m9qigxd7kt0xtta3chrx270grd1l4p.zip",
                    "split_file_type": "zip"
                },
                "AID1843": {
                    "raw_url": "https://vanderbilt.box.com/shared/static/4f9wl7t9pmkm5p6695owj6bsxin0hdo1.zip",
                    "file_type": "zip",
                    "raw_files": ['AID1843_actives.csv', 'AID1843_inactives.csv', 'AID1843_actives.sdf', 'AID1843_inactives.sdf'],
                    "split_url": "https://vanderbilt.box.com/shared/static/o3hmvwo0surg1vhettflhtxxgn2o6clw.zip",
                    "split_file_type": "zip"
                },
                "AID2258": {
                    "raw_url": "https://vanderbilt.box.com/shared/static/b1cg619c4f35p9mkm42mgu8a9m86t4at.zip",
                    "file_type": "zip",
                    "raw_files": ['AID2258_actives.csv', 'AID2258_inactives.csv', 'AID2258_actives.sdf', 'AID2258_inactives.sdf'],
                    "split_url": "https://vanderbilt.box.com/shared/static/2hhz7o3vxwjq6p93yzxfrporeze5zzgf.zip",
                    "split_file_type": "zip"
                },
                "AID2689": {
                    "raw_url": "https://vanderbilt.box.com/shared/static/5no4ut1vusxmxsbwxsmb68o6ix94mxgy.zip",
                    "file_type": "zip",
                    "raw_files": ['AID2689_actives.csv', 'AID2689_inactives.csv', 'AID2689_actives.sdf', 'AID2689_inactives.sdf'],
                    "split_url": "https://vanderbilt.box.com/shared/static/hml9qznmwhhb6rcddw18vp7yoh331wlb.zip",
                    "split_file_type": "zip"
                },
                "AID435008": {
                    "raw_url": "https://vanderbilt.box.com/shared/static/mu6hfitlenanp11wgm3z7u5drae2ofsd.zip",
                    "file_type": "zip",
                    "raw_files": ['AID435008_actives.csv', 'AID435008_inactives.csv', 'AID435008_actives.sdf', 'AID435008_inactives.sdf'],
                    "split_url": "https://vanderbilt.box.com/shared/static/ybyorbb541gnefjvagbqg1p37jfae98z.zip",
                    "split_file_type": "zip"
                },
                "AID435034": {
                    "raw_url": "https://vanderbilt.box.com/shared/static/ctk2vs70bsjmpznalqoj4uqbbdkdcbgj.zip",
                    "file_type": "zip",
                    "raw_files": ['AID435034_actives.csv', 'AID435034_inactives.csv', 'AID435034_actives.sdf', 'AID435034_inactives.sdf'],
                    "split_url": "https://vanderbilt.box.com/shared/static/onistu6vrizl4o2nu0xlfhowcf1q498a.zip",
                    "split_file_type": "zip"
                },
                "AID463087": {
                    "raw_url": "https://vanderbilt.box.com/shared/static/y52n3pzf27ghqtxt0xz899gwuj3yw3m2.zip",
                    "file_type": "zip",
                    "raw_files": ['AID463087_actives.csv', 'AID463087_inactives.csv', 'AID463087_actives.sdf', 'AID463087_inactives.sdf'],
                    "split_url": "https://vanderbilt.box.com/shared/static/y9jkhhu4cljq4awlfkvlo4im6iz3kj5t.zip",
                    "split_file_type": "zip"
                },
                "AID485290": {
                    "raw_url": "https://vanderbilt.box.com/shared/static/8w0njtoqmgs8g1c12qj0p52w0d3wwbop.zip",
                    "file_type": "zip",
                    "raw_files": ['AID485290_actives.csv', 'AID485290_inactives.csv', 'AID485290_actives.sdf', 'AID485290_inactives.sdf'],
                    "split_url": "https://vanderbilt.box.com/shared/static/2w5byzpf5pk4jqb9rknzdihzztb3u9e6.zip",
                    "split_file_type": "zip"
                },
                "AID488997": {
                    "raw_url": "https://vanderbilt.box.com/shared/static/rguy1gm36x7clq6822riconznoflc4xe.zip",
                    "file_type": "zip",
                    "raw_files": ['AID488997_actives.csv', 'AID488997_inactives.csv', 'AID488997_actives.sdf', 'AID488997_inactives.sdf'],
                    "split_url": "https://vanderbilt.box.com/shared/static/dceunwrlotxkz58usfgnxr7meziy3cdr.zip",
                    "split_file_type": "zip"
                }
                }   

    @property
    def raw_file_names(self):
        pattern_csv = os.path.join(self.raw_dir, f'raw_{self.name}_*.csv')
        pattern_sdf = os.path.join(self.raw_dir, f'raw_{self.name}_*.sdf')
        files_csv = glob(pattern_csv)
        files_sdf = glob(pattern_sdf)
        return [os.path.basename(f) for f in files_csv + files_sdf]
    
    @property
    def processed_file_names(self):
        if self.mol_repr == '2d_graph':
            return [f'processed_{self.mol_repr}_{self.source}_{self.name}.pt'] 
        elif self.mol_repr == '3d_graph':
            return [f'processed_{self.mol_repr}_{self.name}.pt'] 
    
    @property
    def processed_dir(self):
        return osp.join(self.root, 'processed', self.mol_repr) 

    def download(self):
        os.makedirs(osp.join(self.root, 'raw'), exist_ok=True)
        os.makedirs(osp.join(self.root, 'split'), exist_ok=True)
        
        dataset_info = self.welqrate_dataset_catalog.get(self.name)
        if not dataset_info:
            print(f"Download information for {self.name} is not available.")
            return

        # Download and extract raw files
        if not all(osp.exists(osp.join(self.root, 'raw', f)) for f in dataset_info['raw_files']):
            print(f"Downloading {self.name} raw files...")
            raw_zip_path = osp.join(self.root, 'raw', f'{self.name}.zip')
            subprocess.run(['curl', '-L', dataset_info['raw_url'], '--output', raw_zip_path], check=True)
            
            print("Extracting raw files...")
            subprocess.run(['unzip', '-j', raw_zip_path, '-d', osp.join(self.root, 'raw')], check=True)
            
            print("Removing raw zip file...")
            os.remove(raw_zip_path)

        # Download and extract split files
        if not os.listdir(osp.join(self.root, 'split')): 
            print(f"Downloading {self.name} split files...")
            split_zip_path = osp.join(self.root, 'split', f'{self.name}_split.zip')
            subprocess.run(['curl', '-L', dataset_info['split_url'], '--output', split_zip_path], check=True)
            
            print("Extracting split files...")
            subprocess.run(['unzip', split_zip_path, '-d', osp.join(self.root, 'split')], check=True)
            
            print("Removing split zip file...")
            os.remove(split_zip_path)

    def process(self):
        print(f'molecule representation: {self.mol_repr}')
        if self.mol_repr == '2d_graph':   
            self.file_type = '.csv'
        elif self.mol_repr == '3d_graph':
            self.file_type = '.sdf'
        print(f'processing dataset {self.name}')
        RDLogger.DisableLog('rdApp.*')
        
        data_list = []
        mol_id = 0
        
        for file_name, label in [(f'{self.name}_actives{self.file_type}', 1),
                                 (f'{self.name}_inactives{self.file_type}', 0)]:

            source_path = os.path.join(self.root, 'raw', file_name)
            print(f'loaded raw file from {source_path}')
            
            if self.mol_repr == '2d_graph':
                inchi_list = pd.read_csv(source_path, sep=',')['InChI'].tolist()
                cid_list = pd.read_csv(source_path, sep=',')['CID'].tolist() 
                smiles_list = pd.read_csv(source_path, sep=',')['SMILES'].tolist()

                for i, _ in tqdm(enumerate(cid_list), total=len(inchi_list)):
                    if self.source == 'inchi':
                        pyg_data = inchi2graph(inchi_list[i])
                    elif self.source == 'smiles':
                        pyg_data = smiles2graph(smiles_list[i])
                    pyg_data.y = torch.tensor([label], dtype=torch.int) 
                    pyg_data.pubchem_cid = torch.tensor(int(cid_list[i]), dtype=torch.int)
                    pyg_data.mol_id = torch.tensor(mol_id, dtype=torch.int)
                    data_list.append(pyg_data)
                    mol_id += 1
                    
            elif self.mol_repr == '3d_graph':
                df = pd.read_csv(f'{source_path[:-4]}.csv')
                smiles_dict = df.set_index('CID')['SMILES'].to_dict()
                inchi_dict = df.set_index('CID')['InChI'].to_dict()

                mol_conformer_list, cid_list = sdffile2mol_conformer(source_path)
                
                for i, (mol, conformer) in tqdm(enumerate(mol_conformer_list), total=len(mol_conformer_list)):
                    pyg_data = mol_conformer2graph3d(mol, conformer)

                    pyg_data.pubchem_cid = torch.tensor([int(cid_list[i])], dtype=torch.int)
                    pyg_data.y = torch.tensor([label], dtype=torch.int)
                    pyg_data.smiles = smiles_dict[int(cid_list[i])]
                    pyg_data.inchi = inchi_dict[int(cid_list[i])]
     
                    pyg_data.mol_id = torch.tensor([mol_id], dtype=torch.int)
                    data_list.append(pyg_data)
                    mol_id += 1
        
        data, slices = self.collate(data_list)
        if self.mol_repr == '2d_graph':
            processed_file_path = os.path.join(self.processed_dir, f'processed_{self.mol_repr}_{self.source}_{self.name}.pt')
        elif self.mol_repr == '3d_graph':
            processed_file_path = os.path.join(self.processed_dir, f'processed_{self.mol_repr}_{self.name}.pt')
        
        torch.save((data, slices), processed_file_path)

        
    def get_idx_split(self, split_scheme = 'random_cv1'):
        
        valid_schemes = [f'random_cv{i}' for i in range(1, 6)] + [f'scaffold_seed{i}' for i in range(1, 6)]
        if split_scheme not in valid_schemes:
            raise ValueError(f"Invalid split_scheme. Must be one of {valid_schemes}")
        
        num = int(split_scheme[-1])
        path = osp.join(self.root, 'split')
        try: 
            if 'random' in split_scheme:
                print(f'loading random split cv{num}')
                split_scheme = f'{self.name}_{self.mol_repr[:2]}_random_cv{num}'
                split_dict = torch.load(osp.join(path, 'random', f'{split_scheme}.pt'))
            elif 'scaffold' in split_scheme:
                print(f'loading scaffold split seed{num}')
                split_scheme = f'{self.name}_{self.mol_repr[:2]}_scaffold_seed{num}'
                split_dict = torch.load(osp.join(path, 'scaffold', f'{split_scheme}.pt'))

        except Exception as e:
            print(f'split file not found. Error msg: {e}')

        # count the number of active molecules in the split
        y = self.y.squeeze()
        num_active_train = len([x for x in y[split_dict['train']] if x == 1])
        print(f'train set: {num_active_train} actives and {len(split_dict["train"]) - num_active_train} inactives')
        num_active_valid = len([x for x in y[split_dict['valid']] if x == 1])
        print(f'valid set: {num_active_valid} actives and {len(split_dict["valid"]) - num_active_valid} inactives')
        num_active_test = len([x for x in y[split_dict['test']] if x == 1])
        print(f'test set: {num_active_test} actives and {len(split_dict["test"]) - num_active_test} inactives')
        
        return split_dict

    def get_dataset_info(self):
        dataset_info = {
            'name': self.name,
            'mol_repr': self.mol_repr,
            'source': self.source,
            'num_active': self.num_active,
            'num_inactive': self.num_inactive
        }
        return dataset_info





import torch
import torch.nn.functional as F
from torch.nn import Linear, Module, ReLU, Dropout
from .GNNConv import GAT as Encoder
from torch_geometric.nn import global_add_pool
import torch


class GAT_Model(torch.nn.Module):
    def __init__(
            self,
            in_channels,
            hidden_channels,
            num_layers,
            heads=1,
            dropout=0.0,
            act='relu',
            one_hot=False,
            act_first=False,
            act_kwargs=None,
            norm=None,
            norm_kwargs=None,
            jk=None
    ):
        super(GAT_Model, self).__init__()

        # Initialize the encoder using the imported GAT class
        self.encoder = Encoder(
            in_channels=in_channels,
            hidden_channels=hidden_channels,
            num_layers=num_layers,
            heads=heads,
            dropout=dropout,
            act=act,
            act_first=act_first,
            act_kwargs=act_kwargs,
            norm=norm,
            norm_kwargs=norm_kwargs,
            jk=jk
        )

        # Pooling operation
        self.pool = global_add_pool
        self.one_hot = one_hot
        self.ffn_dropout = Dropout(p=0.25)
        self.lin1 = Linear(hidden_channels, 64)
        self.lin2 = Linear(64, 1)
        self.activate_func = ReLU()

    def forward(self, batch_data):
        edge_index, edge_attr, batch = batch_data.edge_index, batch_data.edge_attr, batch_data.batch
        if self.one_hot:
            x = batch_data.x_one_hot
            x = x.float()
            node_embedding = self.encoder(x, edge_index, edge_attr=edge_attr)
        else:
            x = batch_data.x
            node_embedding = self.encoder(x, edge_index, edge_attr=edge_attr)
        graph_embedding = self.pool(node_embedding, batch)
        graph_embedding = self.ffn_dropout(graph_embedding)
        # print(graph_embedding.shape)
        prediction = self.lin2(self.activate_func(self.lin1(graph_embedding)))
        return prediction



from torch_geometric.data import Data
from rdkit import Chem
import numpy as np
from .features import (atom_to_feature_vector, atom_to_one_hot_vector,
 bond_to_feature_vector, atom_feature_vector_to_dict, bond_feature_vector_to_dict) 
import rdkit.Chem.rdMolDescriptors as rdMolDescriptors
import rdkit.Chem.EState as EState
import rdkit.Chem.rdPartialCharges as rdPartialCharges
import torch
from torch_geometric.nn import radius_graph



def ReorderCanonicalRankAtoms(mol):
    """
    Reorder the atoms in the molecule based on the canonical rank.
    :param mol: rdkit.Chem.rdchem.Mol
    :return: rdkit.Chem.rdchem.Mol, list
    """
    order = tuple(zip(*sorted([(j, i) for i, j in enumerate(Chem.CanonicalRankAtoms(mol))])))[1]
    mol_renum = Chem.RenumberAtoms(mol, order)
    return mol_renum, order


def atomized_mol_level_features(all_atom_features, mol):
    '''
    Get more atom features that cannot be calculated only with atom,
    but also with mol
    :param all_atom_features:
    :param mol:
    :return:
    '''
    # Crippen has two parts: first is logP, second is Molar Refactivity(MR)
    all_atom_crippen = rdMolDescriptors._CalcCrippenContribs(mol)
    all_atom_TPSA_contrib = rdMolDescriptors._CalcTPSAContribs(mol)
    all_atom_ASA_contrib = rdMolDescriptors._CalcLabuteASAContribs(mol)[0]
    all_atom_EState = EState.EStateIndices(mol)

    new_all_atom_features = []
    for atom_id, feature in enumerate(all_atom_features):
        crippen_logP = all_atom_crippen[atom_id][0]
        crippen_MR = all_atom_crippen[atom_id][1]
        atom_TPSA_contrib = all_atom_TPSA_contrib[atom_id]
        atom_ASA_contrib = all_atom_ASA_contrib[atom_id]
        atom_EState = all_atom_EState[atom_id]

        feature.append(crippen_logP)
        feature.append(crippen_MR)
        feature.append(atom_TPSA_contrib)
        feature.append(atom_ASA_contrib)
        feature.append(atom_EState)

        new_all_atom_features.append(feature)
    return new_all_atom_features


def smiles2graph(smiles_string, removeHs=True, reorder_atoms=False):
    """
    Converts SMILES string to 2D graph Data object
    :input: SMILES string (str)
    :return: graph object
    """
    try:
        mol = Chem.MolFromSmiles(smiles_string)
        mol = Chem.RemoveHs(mol) if removeHs else mol
        if reorder_atoms:
            mol, _ = ReorderCanonicalRankAtoms(mol)

    except Exception as e:
        print(f'cannot generate mol, error: {e}, smiles: {smiles_string}')

    if mol is None:
        smiles = smiles_cleaner(smiles_string)
        try:
            mol = Chem.MolFromSmiles(smiles)
            mol = Chem.RemoveHs(mol) if removeHs else mol
            if reorder_atoms:
                mol, _ = ReorderCanonicalRankAtoms(mol)

        except Exception as e:
            print(f'cannot generate mol, error: {e}, smiles: {smiles_string}')
            mol = None

        if mol is None:
            raise ValueError(f'cannot generate molecule with smiles: {smiles_string}')
    else:
        # calculate Gasteiger charges
        rdPartialCharges.ComputeGasteigerCharges(mol)
        
        # atoms
        atom_features_list = []
        # atom_num_list = []
        one_hot_atom_list = []
        for atom in mol.GetAtoms():
            # atom_num_list.append(atom.GetAtomicNum())
            atom_features_list.append(atom_to_feature_vector(atom))
            one_hot_atom_list.append(atom_to_one_hot_vector(atom))

        atom_features_list = atomized_mol_level_features(atom_features_list, mol)
        x = torch.tensor(atom_features_list, dtype=torch.float32)
        # atom_num = torch.tensor(atom_num_list, dtype=torch.int64)
        one_hot_atom = torch.tensor(one_hot_atom_list, dtype=torch.int64)

        # bonds
        if len(mol.GetBonds()) > 0:  # mol has bonds
            edges_list = []
            edge_features_list = []
            for bond in mol.GetBonds():
                i = bond.GetBeginAtomIdx()
                j = bond.GetEndAtomIdx()
                
                bond_attr = []
                bond_attr += one_hot_vector(bond.GetBondTypeAsDouble(),
                                            [1.0, 1.5, 2.0, 3.0])
                is_aromatic = bond.GetIsAromatic()
                is_conjugated = bond.GetIsConjugated()
                is_in_ring = bond.IsInRing()
                bond_attr.append(is_aromatic)
                bond_attr.append(is_conjugated)
                bond_attr.append(is_in_ring)

                # add edges in both directions
                edges_list.append((i, j))
                edge_features_list.append(bond_attr)
                edges_list.append((j, i))
                edge_features_list.append(bond_attr)

            edge_index = torch.tensor(edges_list, dtype=torch.long).t().contiguous()
            edge_attr = torch.tensor(edge_features_list, dtype=torch.float32)

        else:  # mol has no bonds
            print('molecule does not have bond:', smiles_string)
            edge_index = torch.empty((2, 0), dtype=torch.long)
            edge_attr = torch.empty((0, 7), dtype=torch.float32)

        graph = Data(
            edge_index=edge_index,
            edge_attr=edge_attr,
            x=x,
            x_one_hot=one_hot_atom,
            # atom_num=atom_num,
            num_nodes=torch.tensor([len(x)]),
            num_edges=len(edge_index[0]),
            smiles=smiles_string
        )

    return graph


def inchi2graph(inchi_string, removeHs=True, reorder_atoms=False):
    """
    Converts Inchi string to 2D graph Data object
    :input: SMILES string (str)
    :return: graph object
    """
    try:
        mol = Chem.MolFromInchi(inchi_string)
        mol = Chem.RemoveHs(mol) if removeHs else mol
        if reorder_atoms:
            mol, _ = ReorderCanonicalRankAtoms(mol)

    except Exception as e:
        print(f'cannot generate mol, error: {e}, inchi: {inchi_string}')
        mol = None

    if mol is None:
        raise ValueError(f'cannot generate molecule with inchi: {inchi_string}')

    else:
        # calculate Gasteiger charges
        rdPartialCharges.ComputeGasteigerCharges(mol)
        
        # atoms
        atom_features_list = []
        # atom_num_list = []
        one_hot_atom_list = []
        for atom in mol.GetAtoms():
            # atom_num_list.append(atom.GetAtomicNum())
            atom_features_list.append(atom_to_feature_vector(atom))
            one_hot_atom_list.append(atom_to_one_hot_vector(atom))

        atom_features_list = atomized_mol_level_features(atom_features_list, mol)
        x = torch.tensor(atom_features_list, dtype=torch.float32)
        # atom_num = torch.tensor(atom_num_list, dtype=torch.int64)
        one_hot_atom = torch.tensor(one_hot_atom_list, dtype=torch.int64)

        # bond features: bond type, bond stereo, is_conjugated
        if len(mol.GetBonds()) > 0:  # mol has bonds
            edges_list = []
            edge_features_list = []
            for bond in mol.GetBonds():
                i = bond.GetBeginAtomIdx()
                j = bond.GetEndAtomIdx()
                
                bond_attr = []
                bond_attr += one_hot_vector(bond.GetBondTypeAsDouble(),
                                            [1.0, 1.5, 2.0, 3.0])
                is_aromatic = bond.GetIsAromatic()
                is_conjugated = bond.GetIsConjugated()
                is_in_ring = bond.IsInRing()
                bond_attr.append(is_aromatic)
                bond_attr.append(is_conjugated)
                bond_attr.append(is_in_ring)

                # add edges in both directions
                edges_list.append((i, j))
                edge_features_list.append(bond_attr)
                edges_list.append((j, i))
                edge_features_list.append(bond_attr)

            edge_index = torch.tensor(edges_list, dtype = torch.long).t().contiguous()
            edge_attr = torch.tensor(edge_features_list, dtype=torch.float32)

        else: 
            print('molecule does not have bond:', inchi_string)
            edge_index = torch.empty((2, 0), dtype=torch.long)
            edge_attr = torch.empty((0, 7), dtype=torch.float32)

        graph = Data(
            edge_index=edge_index,
            edge_attr=edge_attr,
            x=x,
            x_one_hot=one_hot_atom,
            # atom_num=atom_num,
            num_nodes=torch.tensor([len(x)]),
            num_edges=len(edge_index[0]),
            inchi=inchi_string,
        )

    return graph
    

# Functions that convert sdf files to molecular graphs
def sdffile2graph3d_lst(sdffile):
    """convert SDF file into a list of 3D graph.

    Args:
      sdffile: SDF file

    Returns:
      graph3d_lst: a list of 3D graph.
            each graph has (i) idx2atom (dict); (ii) distance_adj_matrix (np.array); (iii) bondtype_adj_matrix (np.array)

    """
    mol_conformer_lst = sdffile2mol_conformer(sdffile)
    graph3d_lst = mol_conformer2graph3d(mol_conformer_lst)
    return graph3d_lst

def sdffile2mol_conformer(sdffile):
    """convert sdffile into a list of molecule conformers.
    Args:
      sdffile: str, file
    Returns:
      mol_conformer_lst: a list of (molecule, conformer). 
      cid_lst: a list of pubchem cid of the molecules
    """

    supplier = Chem.SDMolSupplier(sdffile, removeHs=True, sanitize=False)
    mol_lst = [mol for mol in supplier if mol is not None]
    conformer_lst = []
    cid_lst = []
    for mol in mol_lst:
        if mol is not None:
            conformer = mol.GetConformer(id=0)
            cid = mol.GetProp('_Name')
            conformer_lst.append(conformer)
            cid_lst.append(cid)
    mol_conformer_lst = list(zip(mol_lst, conformer_lst))
    return mol_conformer_lst, cid_lst


def mol_conformer2graph3d(mol, conformer, removeHs=True, reorder_atoms=False,
                          r=6.0, max_num_neighbors=50):
    """convert (molecule, conformer) into a 3D graph.
    Args:
      mol_conformer: tuple (molecule, conformer)
      r: radius of the graph
      max_num_neighbors: maximum number of neighbors
    Returns:
      graph3d: a 3D pyg graph with 3D coordinates
    """
    if mol is None:
        raise ValueError(f'cannot generate molecule with conformer: {conformer}')
    else:
        mol = Chem.RemoveHs(mol, sanitize=False) if removeHs else mol
        if reorder_atoms:
            mol, _ = ReorderCanonicalRankAtoms(mol)  

        # calculate Gasteiger charges
        rdPartialCharges.ComputeGasteigerCharges(mol)
        
        # atoms
        atom_features_list = []
        atom_num_list = []
        one_hot_atom_list = []

        for atom in mol.GetAtoms():
            atom_num_list.append(atom.GetAtomicNum())
            atom_features_list.append(atom_to_feature_vector(atom))
            one_hot_atom_list.append(atom_to_one_hot_vector(atom))

        atom_features_list = atomized_mol_level_features(atom_features_list, mol)
        x = torch.tensor(atom_features_list, dtype=torch.float32)
        atom_num = torch.tensor(atom_num_list, dtype=torch.int64)
        one_hot_atom = torch.tensor(one_hot_atom_list, dtype=torch.int64)

        positions = []
        for i in range(len(atom_num)):
            pos = conformer.GetAtomPosition(i)
            coordinate = np.array([pos.x, pos.y, pos.z]).reshape(1, 3)
            positions.append(coordinate)
        positions = np.concatenate(positions, 0)
        
        # edge
        edge_index = radius_graph(torch.tensor(positions), r=r, max_num_neighbors=max_num_neighbors).to(torch.long)
                    
        graph = Data(
            x = x,
            edge_index = edge_index,
            pos = torch.tensor(positions, dtype=torch.float32),
            x_one_hot = one_hot_atom,
            atom_num = atom_num,
            num_edges = len(edge_index[0]),
            num_nodes = torch.tensor([len(x)]),
        )
    
    return graph

pattern_dict = {'[NH-]': '[N-]', '[OH2+]':'[O]'}

def smiles_cleaner(smiles):
    '''
    This function is to clean smiles for some known issues that makes
    rdkit:Chem.MolFromSmiles not working
    '''
    print('fixing smiles for rdkit...')
    new_smiles = smiles
    for pattern, replace_value in pattern_dict.items():
        if pattern in smiles:
            print('found pattern and fixed the smiles!')
            new_smiles = smiles.replace(pattern, replace_value)
    return new_smiles

def one_hot_vector(val, lst):
	'''
	Converts a value to a one-hot vector based on options in lst
	'''
	if val not in lst:
		val = lst[-1]
	return map(lambda x: x == val, lst)


### Test
if __name__ == '__main__':
    graph = smiles2graph('O1C=C[C@H]([C@H]1O2)c3c2cc(OC)c4c3OC(=O)C5=C4CCC(=O)5')

    # load sdf file
    sdf_file = 'test.sdf'
    # sdf_supplier = Chem.SDMolSupplier(sdf_file)
    # print(len(sdf_supplier))
    mol_conformer_lst = sdffile2mol_conformer(sdf_file)
    for mol, conformer in mol_conformer_lst:
        
        data = mol_conformer2graph3d(mol, conformer)
        print(data)
        print(data.x[:5, :])
        print(data.pos[:5, :])
        print(data.edge_attr[:5, :]) 

        break
    

    



import pandas as pd
import os
import csv
import copy
import torch
from welqrate.train import train


def save_summary_statistics(seed_results, results_dir, dataset_name, split_scheme, timestamp):
    """
    Calculate and save summary statistics across multiple seeds.
    
    Args:
        seed_results: List of dictionaries containing results from different seeds
        results_dir: Directory to save results
        dataset_name: Name of the dataset
        split_scheme: Split scheme used
        timestamp: Timestamp for file naming
    """
    if not seed_results:
        print("No valid seed results to summarize")
        return
        
    seed_df = pd.DataFrame(seed_results)
    print("\nResults across seeds:")
    # Print summary statistics
    print(f"Mean test logAUC: {seed_df['test_logAUC'].mean():.4f} ± {seed_df['test_logAUC'].std():.4f}")
    print(f"Mean test EF100: {seed_df['test_EF100'].mean():.4f} ± {seed_df['test_EF100'].std():.4f}")
    print(f"Mean test DCG100: {seed_df['test_DCG100'].mean():.4f} ± {seed_df['test_DCG100'].std():.4f}")
    print(f"Mean test BEDROC: {seed_df['test_BEDROC'].mean():.4f} ± {seed_df['test_BEDROC'].std():.4f}")
    print(f"Mean test EF500: {seed_df['test_EF500'].mean():.4f} ± {seed_df['test_EF500'].std():.4f}")
    print(f"Mean test EF1000: {seed_df['test_EF1000'].mean():.4f} ± {seed_df['test_EF1000'].std():.4f}")
    print(f"Mean test DCG500: {seed_df['test_DCG500'].mean():.4f} ± {seed_df['test_DCG500'].std():.4f}")
    print(f"Mean test DCG1000: {seed_df['test_DCG1000'].mean():.4f} ± {seed_df['test_DCG1000'].std():.4f}")

    # Save summary statistics to CSV
    summary_stats = {
        'Metric': ['logAUC', 'EF100', 'DCG100', 'BEDROC', 'EF500', 'EF1000', 'DCG500', 'DCG1000'],
        'Mean': [
            seed_df['test_logAUC'].mean(),
            seed_df['test_EF100'].mean(),
            seed_df['test_DCG100'].mean(),
            seed_df['test_BEDROC'].mean(),
            seed_df['test_EF500'].mean(),
            seed_df['test_EF1000'].mean(),
            seed_df['test_DCG500'].mean(),
            seed_df['test_DCG1000'].mean()
        ],
        'Std': [
            seed_df['test_logAUC'].std(),
            seed_df['test_EF100'].std(),
            seed_df['test_DCG100'].std(),
            seed_df['test_BEDROC'].std(),
            seed_df['test_EF500'].std(),
            seed_df['test_EF1000'].std(),
            seed_df['test_DCG500'].std(),
            seed_df['test_DCG1000'].std()
        ]
    }
    summary_df = pd.DataFrame(summary_stats)
    summary_csv = f'{results_dir}/gcn_summary_stats_{dataset_name}_{split_scheme}_{timestamp}.csv'
    summary_df.to_csv(summary_csv, index=False)


def run_with_seed(seed, model, best_params, base_config, args, train_loader, valid_loader, test_loader, device, results_dir, csv_file):
    print(f"\nRunning with seed {seed}")
    config = copy.deepcopy(base_config)
    config['general']['seed'] = seed
    config['train']['peak_lr'] = best_params['peak_lr']
    config['data']['split_scheme'] = args.split
    config['model']['hidden_channels'] = best_params['hidden_channels']
    config['model']['num_layers'] = best_params['num_layers']
    config['train']['num_epochs'] = args.num_epochs

    final_results_dir = f'{results_dir}/{args.dataset}/{args.split}/gcn/seed{seed}'
    os.makedirs(final_results_dir, exist_ok=True)
    
    try:
        # Train model and get metrics
        test_logAUC, test_EF100, test_DCG100, test_BEDROC, test_EF500, test_EF1000, test_DCG500, test_DCG1000 = train(
            model=model, 
            train_loader=train_loader,
            valid_loader=valid_loader,
            test_loader=test_loader,
            config=config, 
            device=device,
            save_path=final_results_dir,
        )
        
        result = {
            'seed': seed,
            'test_logAUC': test_logAUC,
            'test_EF100': test_EF100, 
            'test_DCG100': test_DCG100,
            'test_BEDROC': test_BEDROC,
            'test_EF500': test_EF500,
            'test_EF1000': test_EF1000,
            'test_DCG500': test_DCG500,
            'test_DCG1000': test_DCG1000
        }

        # Save seed results to CSV
        with open(csv_file, 'a', newline='') as f:
            writer = csv.writer(f)
            writer.writerow([
                best_params['hidden_channels'],
                best_params['num_layers'],
                best_params['peak_lr'],
                test_logAUC,
                test_EF100, 
                test_DCG100,
                test_BEDROC,
                test_EF500,
                test_EF1000,
                test_DCG500,
                test_DCG1000,
                seed
            ])
        
        return result

    except Exception as e:
        print(f"Error occurred with seed {seed}")
        print(f"Error message: {str(e)}")
        return None





# Open the input file for reading

def rank_prediction(type, model_name):
    with open(f'result/{model_name}/per_molecule_pred_of_{type}_set.txt', 'r') as f:
        data = [(float(line.split('\t')[0]), line.split('\t')[1] ) for line in
                f.readlines()]

    ranked_data = sorted(data, key=lambda x: x[0], reverse=True)

    with open(f'result/{model_name}/ranked_mol_score_{type}.txt', 'w') as f:
        for i, (score, label) in enumerate(ranked_data):
            f.write(f"{i}\t{score}\t{label}")


import copy
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch
import torch.nn.functional as F
from torch import Tensor
from torch.nn import Linear, ModuleList
from tqdm import tqdm

from torch_geometric.loader import NeighborLoader
from torch_geometric.nn.conv import (
    EdgeConv,
    GATConv,
    GATv2Conv,
    GCNConv,
    GINConv,
    MessagePassing,
    PNAConv,
    SAGEConv,
)
from torch_geometric.nn.models import MLP
from torch_geometric.nn.models.jumping_knowledge import JumpingKnowledge
from torch_geometric.nn.resolver import (
    activation_resolver,
    normalization_resolver,
)
from torch_geometric.typing import Adj, OptTensor


class BasicGNN(torch.nn.Module):
    r"""An abstract class for implementing basic GNN models.

    Args:
        in_channels (int or tuple): Size of each input sample, or :obj:`-1` to
            derive the size from the first input(s) to the forward method.
            A tuple corresponds to the sizes of source and target
            dimensionalities.
        hidden_channels (int): Size of each hidden sample.
        num_layers (int): Number of message passing layers.
        out_channels (int, optional): If not set to :obj:`None`, will apply a
            final linear transformation to convert hidden node embeddings to
            output size :obj:`out_channels`. (default: :obj:`None`)
        dropout (float, optional): Dropout probability. (default: :obj:`0.`)
        act (str or Callable, optional): The non-linear activation function to
            use. (default: :obj:`"relu"`)
        act_first (bool, optional): If set to :obj:`True`, activation is
            applied before normalization. (default: :obj:`False`)
        act_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective activation function defined by :obj:`act`.
            (default: :obj:`None`)
        norm (str or Callable, optional): The normalization function to
            use. (default: :obj:`None`)
        norm_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective normalization function defined by :obj:`norm`.
            (default: :obj:`None`)
        jk (str, optional): The Jumping Knowledge mode. If specified, the model
            will additionally apply a final linear transformation to transform
            node embeddings to the expected output feature dimensionality.
            (:obj:`None`, :obj:`"last"`, :obj:`"cat"`, :obj:`"max"`,
            :obj:`"lstm"`). (default: :obj:`None`)
        **kwargs (optional): Additional arguments of the underlying
            :class:`torch_geometric.nn.conv.MessagePassing` layers.
    """
    def __init__(
        self,
        in_channels: int,
        hidden_channels: int,
        num_layers: int,
        out_channels: Optional[int] = None,
        dropout: float = 0.0,
        act: Union[str, Callable, None] = "relu",
        act_first: bool = False,
        act_kwargs: Optional[Dict[str, Any]] = None,
        norm: Union[str, Callable, None] = None,
        norm_kwargs: Optional[Dict[str, Any]] = None,
        jk: Optional[str] = None,
        **kwargs,
    ):
        super().__init__()

        self.in_channels = in_channels
        self.hidden_channels = hidden_channels
        self.num_layers = num_layers

        self.dropout = dropout
        self.act = activation_resolver(act, **(act_kwargs or {}))
        self.jk_mode = jk
        self.act_first = act_first
        self.norm = norm if isinstance(norm, str) else None
        self.norm_kwargs = norm_kwargs

        if out_channels is not None:
            self.out_channels = out_channels
        else:
            self.out_channels = hidden_channels

        self.convs = ModuleList()
        if num_layers > 1:
            self.convs.append(
                self.init_conv(in_channels, hidden_channels, **kwargs))
            if isinstance(in_channels, (tuple, list)):
                in_channels = (hidden_channels, hidden_channels)
            else:
                in_channels = hidden_channels
        for _ in range(num_layers - 2):
            self.convs.append(
                self.init_conv(in_channels, hidden_channels, **kwargs))
            if isinstance(in_channels, (tuple, list)):
                in_channels = (hidden_channels, hidden_channels)
            else:
                in_channels = hidden_channels
        if out_channels is not None and jk is None:
            self._is_conv_to_out = True
            self.convs.append(
                self.init_conv(in_channels, out_channels, **kwargs))
        else:
            self.convs.append(
                self.init_conv(in_channels, hidden_channels, **kwargs))

        self.norms = None

        if norm is not None:
            norm_layer = normalization_resolver(
                norm,
                hidden_channels,
                **(norm_kwargs or {}),
            )
            self.norms = ModuleList()
            for _ in range(num_layers - 1):
                self.norms.append(copy.deepcopy(norm_layer))
            if jk is not None:
                self.norms.append(copy.deepcopy(norm_layer))

        if jk is not None and jk != 'last':
            self.jk = JumpingKnowledge(jk, hidden_channels, num_layers)

        if jk is not None:
            if jk == 'cat':
                in_channels = num_layers * hidden_channels
            else:
                in_channels = hidden_channels
            self.lin = Linear(in_channels, self.out_channels)


    def init_conv(self, in_channels: Union[int, Tuple[int, int]],
                  out_channels: int, **kwargs) -> MessagePassing:
        raise NotImplementedError

    def reset_parameters(self):
        r"""Resets all learnable parameters of the module."""
        for conv in self.convs:
            conv.reset_parameters()
        for norm in self.norms or []:
            norm.reset_parameters()
        if hasattr(self, 'jk'):
            self.jk.reset_parameters()
        if hasattr(self, 'lin'):
            self.lin.reset_parameters()

    def forward(
        self,
        x: Tensor,
        edge_index: Adj,
        *,
        edge_weight: OptTensor = None,
        edge_attr: OptTensor = None,
    ) -> Tensor:
        r"""
        Args:
            x (torch.Tensor): The input node features.
            edge_index (torch.Tensor): The edge indices.
            edge_weight (torch.Tensor, optional): The edge weights (if
                supported by the underlying GNN layer). (default: :obj:`None`)
            edge_attr (torch.Tensor, optional): The edge features (if supported
                by the underlying GNN layer). (default: :obj:`None`)
        """
        xs: List[Tensor] = []
        for i in range(self.num_layers):
            # Tracing the module is not allowed with *args and **kwargs :(
            # As such, we rely on a static solution to pass optional edge
            # weights and edge attributes to the module.
            if self.supports_edge_weight and self.supports_edge_attr:
                x = self.convs[i](x, edge_index, edge_weight=edge_weight,
                                  edge_attr=edge_attr)
            elif self.supports_edge_weight:
                x = self.convs[i](x, edge_index, edge_weight=edge_weight)
            elif self.supports_edge_attr:
                x = self.convs[i](x, edge_index, edge_attr=edge_attr)
            else:
                x = self.convs[i](x, edge_index)
            if i == self.num_layers - 1 and self.jk_mode is None:
                break
            if self.act is not None and self.act_first:
                x = self.act(x)
            if self.norms is not None:
                x = self.norms[i](x)
            if self.act is not None and not self.act_first:
                x = self.act(x)
            x = F.dropout(x, p=self.dropout, training=self.training)
            if hasattr(self, 'jk'):
                xs.append(x)

        x = self.jk(xs) if hasattr(self, 'jk') else x
        x = self.lin(x) if hasattr(self, 'lin') else x
        return x

    @torch.no_grad()
    def inference(self, loader: NeighborLoader,
                  device: Optional[torch.device] = None,
                  progress_bar: bool = False) -> Tensor:
        r"""Performs layer-wise inference on large-graphs using a
        :class:`~torch_geometric.loader.NeighborLoader`, where
        :class:`~torch_geometric.loader.NeighborLoader` should sample the
        full neighborhood for only one layer.
        This is an efficient way to compute the output embeddings for all
        nodes in the graph.
        Only applicable in case :obj:`jk=None` or `jk='last'`.
        """
        assert self.jk_mode is None or self.jk_mode == 'last'
        assert isinstance(loader, NeighborLoader)
        assert len(loader.dataset) == loader.data.num_nodes
        assert len(loader.node_sampler.num_neighbors) == 1
        assert not self.training
        # assert not loader.shuffle  # TODO (matthias) does not work :(
        if progress_bar:
            pbar = tqdm(total=len(self.convs) * len(loader))
            pbar.set_description('Inference')

        x_all = loader.data.x.cpu()
        loader.data.n_id = torch.arange(x_all.size(0))

        for i in range(self.num_layers):
            xs: List[Tensor] = []
            for batch in loader:
                x = x_all[batch.n_id].to(device)
                if hasattr(batch, 'adj_t'):
                    edge_index = batch.adj_t.to(device)
                else:
                    edge_index = batch.edge_index.to(device)
                x = self.convs[i](x, edge_index)[:batch.batch_size]
                if i == self.num_layers - 1 and self.jk_mode is None:
                    xs.append(x.cpu())
                    if progress_bar:
                        pbar.update(1)
                    continue
                if self.act is not None and self.act_first:
                    x = self.act(x)
                if self.norms is not None:
                    x = self.norms[i](x)
                if self.act is not None and not self.act_first:
                    x = self.act(x)
                if i == self.num_layers - 1 and hasattr(self, 'lin'):
                    x = self.lin(x)
                xs.append(x.cpu())
                if progress_bar:
                    pbar.update(1)
            x_all = torch.cat(xs, dim=0)
        if progress_bar:
            pbar.close()
        del loader.data.n_id

        return x_all

    def __repr__(self) -> str:
        return (f'{self.__class__.__name__}({self.in_channels}, '
                f'{self.out_channels}, num_layers={self.num_layers})')


class GCN(BasicGNN):
    r"""The Graph Neural Network from the `"Semi-supervised
    Classification with Graph Convolutional Networks"
    <https://arxiv.org/abs/1609.02907>`_ paper, using the
    :class:`~torch_geometric.nn.conv.GCNConv` operator for message passing.

    Args:
        in_channels (int): Size of each input sample, or :obj:`-1` to derive
            the size from the first input(s) to the forward method.
        hidden_channels (int): Size of each hidden sample.
        num_layers (int): Number of message passing layers.
        out_channels (int, optional): If not set to :obj:`None`, will apply a
            final linear transformation to convert hidden node embeddings to
            output size :obj:`out_channels`. (default: :obj:`None`)
        dropout (float, optional): Dropout probability. (default: :obj:`0.`)
        act (str or Callable, optional): The non-linear activation function to
            use. (default: :obj:`"relu"`)
        act_first (bool, optional): If set to :obj:`True`, activation is
            applied before normalization. (default: :obj:`False`)
        act_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective activation function defined by :obj:`act`.
            (default: :obj:`None`)
        norm (str or Callable, optional): The normalization function to
            use. (default: :obj:`None`)
        norm_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective normalization function defined by :obj:`norm`.
            (default: :obj:`None`)
        jk (str, optional): The Jumping Knowledge mode. If specified, the model
            will additionally apply a final linear transformation to transform
            node embeddings to the expected output feature dimensionality.
            (:obj:`None`, :obj:`"last"`, :obj:`"cat"`, :obj:`"max"`,
            :obj:`"lstm"`). (default: :obj:`None`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.GCNConv`.
    """
    supports_edge_weight = False
    supports_edge_attr = False

    def init_conv(self, in_channels: int, out_channels: int,
                  **kwargs) -> MessagePassing:
        return GCNConv(in_channels, out_channels, **kwargs)



class GraphSAGE(BasicGNN):
    r"""The Graph Neural Network from the `"Inductive Representation Learning
    on Large Graphs" <https://arxiv.org/abs/1706.02216>`_ paper, using the
    :class:`~torch_geometric.nn.SAGEConv` operator for message passing.

    Args:
        in_channels (int or tuple): Size of each input sample, or :obj:`-1` to
            derive the size from the first input(s) to the forward method.
            A tuple corresponds to the sizes of source and target
            dimensionalities.
        hidden_channels (int): Size of each hidden sample.
        num_layers (int): Number of message passing layers.
        out_channels (int, optional): If not set to :obj:`None`, will apply a
            final linear transformation to convert hidden node embeddings to
            output size :obj:`out_channels`. (default: :obj:`None`)
        dropout (float, optional): Dropout probability. (default: :obj:`0.`)
        act (str or Callable, optional): The non-linear activation function to
            use. (default: :obj:`"relu"`)
        act_first (bool, optional): If set to :obj:`True`, activation is
            applied before normalization. (default: :obj:`False`)
        act_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective activation function defined by :obj:`act`.
            (default: :obj:`None`)
        norm (str or Callable, optional): The normalization function to
            use. (default: :obj:`None`)
        norm_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective normalization function defined by :obj:`norm`.
            (default: :obj:`None`)
        jk (str, optional): The Jumping Knowledge mode. If specified, the model
            will additionally apply a final linear transformation to transform
            node embeddings to the expected output feature dimensionality.
            (:obj:`None`, :obj:`"last"`, :obj:`"cat"`, :obj:`"max"`,
            :obj:`"lstm"`). (default: :obj:`None`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.SAGEConv`.
    """
    supports_edge_weight = False
    supports_edge_attr = True

    def init_conv(self, in_channels: Union[int, Tuple[int, int]],
                  out_channels: int, **kwargs) -> MessagePassing:
        return SAGEConv(in_channels, out_channels, **kwargs)



class GIN(BasicGNN):
    r"""The Graph Neural Network from the `"How Powerful are Graph Neural
    Networks?" <https://arxiv.org/abs/1810.00826>`_ paper, using the
    :class:`~torch_geometric.nn.GINConv` operator for message passing.

    Args:
        in_channels (int): Size of each input sample.
        hidden_channels (int): Size of each hidden sample.
        num_layers (int): Number of message passing layers.
        out_channels (int, optional): If not set to :obj:`None`, will apply a
            final linear transformation to convert hidden node embeddings to
            output size :obj:`out_channels`. (default: :obj:`None`)
        dropout (float, optional): Dropout probability. (default: :obj:`0.`)
        act (str or Callable, optional): The non-linear activation function to
            use. (default: :obj:`"relu"`)
        act_first (bool, optional): If set to :obj:`True`, activation is
            applied before normalization. (default: :obj:`False`)
        act_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective activation function defined by :obj:`act`.
            (default: :obj:`None`)
        norm (str or Callable, optional): The normalization function to
            use. (default: :obj:`None`)
        norm_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective normalization function defined by :obj:`norm`.
            (default: :obj:`None`)
        jk (str, optional): The Jumping Knowledge mode. If specified, the model
            will additionally apply a final linear transformation to transform
            node embeddings to the expected output feature dimensionality.
            (:obj:`None`, :obj:`"last"`, :obj:`"cat"`, :obj:`"max"`,
            :obj:`"lstm"`). (default: :obj:`None`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.GINConv`.
    """
    supports_edge_weight = False
    supports_edge_attr = False

    def init_conv(self, in_channels: int, out_channels: int,
                  **kwargs) -> MessagePassing:
        mlp = MLP(
            [in_channels, out_channels, out_channels],
            act=self.act,
            act_first=self.act_first,
            norm=self.norm,
            norm_kwargs=self.norm_kwargs,
        )
        return GINConv(mlp, **kwargs)



class GAT(BasicGNN):
    r"""The Graph Neural Network from `"Graph Attention Networks"
    <https://arxiv.org/abs/1710.10903>`_ or `"How Attentive are Graph Attention
    Networks?" <https://arxiv.org/abs/2105.14491>`_ papers, using the
    :class:`~torch_geometric.nn.GATConv` or
    :class:`~torch_geometric.nn.GATv2Conv` operator for message passing,
    respectively.

    Args:
        in_channels (int or tuple): Size of each input sample, or :obj:`-1` to
            derive the size from the first input(s) to the forward method.
            A tuple corresponds to the sizes of source and target
            dimensionalities.
        hidden_channels (int): Size of each hidden sample.
        num_layers (int): Number of message passing layers.
        out_channels (int, optional): If not set to :obj:`None`, will apply a
            final linear transformation to convert hidden node embeddings to
            output size :obj:`out_channels`. (default: :obj:`None`)
        v2 (bool, optional): If set to :obj:`True`, will make use of
            :class:`~torch_geometric.nn.conv.GATv2Conv` rather than
            :class:`~torch_geometric.nn.conv.GATConv`. (default: :obj:`False`)
        dropout (float, optional): Dropout probability. (default: :obj:`0.`)
        act (str or Callable, optional): The non-linear activation function to
            use. (default: :obj:`"relu"`)
        act_first (bool, optional): If set to :obj:`True`, activation is
            applied before normalization. (default: :obj:`False`)
        act_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective activation function defined by :obj:`act`.
            (default: :obj:`None`)
        norm (str or Callable, optional): The normalization function to
            use. (default: :obj:`None`)
        norm_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective normalization function defined by :obj:`norm`.
            (default: :obj:`None`)
        jk (str, optional): The Jumping Knowledge mode. If specified, the model
            will additionally apply a final linear transformation to transform
            node embeddings to the expected output feature dimensionality.
            (:obj:`None`, :obj:`"last"`, :obj:`"cat"`, :obj:`"max"`,
            :obj:`"lstm"`). (default: :obj:`None`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.GATConv` or
            :class:`torch_geometric.nn.conv.GATv2Conv`.
    """
    supports_edge_weight = False
    supports_edge_attr = True

    def init_conv(self, in_channels: Union[int, Tuple[int, int]],
                  out_channels: int, **kwargs) -> MessagePassing:

        v2 = kwargs.pop('v2', False)
        heads = kwargs.pop('heads', 1)
        concat = kwargs.pop('concat', True)

        # Do not use concatenation in case the layer `GATConv` layer maps to
        # the desired output channels (out_channels != None and jk != None):
        if getattr(self, '_is_conv_to_out', False):
            concat = False

        if concat and out_channels % heads != 0:
            raise ValueError(f"Ensure that the number of output channels of "
                             f"'GATConv' (got '{out_channels}') is divisible "
                             f"by the number of heads (got '{heads}')")

        if concat:
            out_channels = out_channels // heads

        Conv = GATConv if not v2 else GATv2Conv
        return Conv(in_channels, out_channels, heads=heads, concat=concat,
                    dropout=self.dropout, **kwargs)



class PNA(BasicGNN):
    r"""The Graph Neural Network from the `"Principal Neighbourhood Aggregation
    for Graph Nets" <https://arxiv.org/abs/2004.05718>`_ paper, using the
    :class:`~torch_geometric.nn.conv.PNAConv` operator for message passing.

    Args:
        in_channels (int): Size of each input sample, or :obj:`-1` to derive
            the size from the first input(s) to the forward method.
        hidden_channels (int): Size of each hidden sample.
        num_layers (int): Number of message passing layers.
        out_channels (int, optional): If not set to :obj:`None`, will apply a
            final linear transformation to convert hidden node embeddings to
            output size :obj:`out_channels`. (default: :obj:`None`)
        dropout (float, optional): Dropout probability. (default: :obj:`0.`)
        act (str or Callable, optional): The non-linear activation function to
            use. (default: :obj:`"relu"`)
        act_first (bool, optional): If set to :obj:`True`, activation is
            applied before normalization. (default: :obj:`False`)
        act_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective activation function defined by :obj:`act`.
            (default: :obj:`None`)
        norm (str or Callable, optional): The normalization function to
            use. (default: :obj:`None`)
        norm_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective normalization function defined by :obj:`norm`.
            (default: :obj:`None`)
        jk (str, optional): The Jumping Knowledge mode. If specified, the model
            will additionally apply a final linear transformation to transform
            node embeddings to the expected output feature dimensionality.
            (:obj:`None`, :obj:`"last"`, :obj:`"cat"`, :obj:`"max"`,
            :obj:`"lstm"`). (default: :obj:`None`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.PNAConv`.
    """
    supports_edge_weight = False
    supports_edge_attr = True

    def init_conv(self, in_channels: int, out_channels: int,
                  **kwargs) -> MessagePassing:
        return PNAConv(in_channels, out_channels, **kwargs)



class EdgeCNN(BasicGNN):
    r"""The Graph Neural Network from the `"Dynamic Graph CNN for Learning on
    Point Clouds" <https://arxiv.org/abs/1801.07829>`_ paper, using the
    :class:`~torch_geometric.nn.conv.EdgeConv` operator for message passing.

    Args:
        in_channels (int): Size of each input sample.
        hidden_channels (int): Size of each hidden sample.
        num_layers (int): Number of message passing layers.
        out_channels (int, optional): If not set to :obj:`None`, will apply a
            final linear transformation to convert hidden node embeddings to
            output size :obj:`out_channels`. (default: :obj:`None`)
        dropout (float, optional): Dropout probability. (default: :obj:`0.`)
        act (str or Callable, optional): The non-linear activation function to
            use. (default: :obj:`"relu"`)
        act_first (bool, optional): If set to :obj:`True`, activation is
            applied before normalization. (default: :obj:`False`)
        act_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective activation function defined by :obj:`act`.
            (default: :obj:`None`)
        norm (str or Callable, optional): The normalization function to
            use. (default: :obj:`None`)
        norm_kwargs (Dict[str, Any], optional): Arguments passed to the
            respective normalization function defined by :obj:`norm`.
            (default: :obj:`None`)
        jk (str, optional): The Jumping Knowledge mode. If specified, the model
            will additionally apply a final linear transformation to transform
            node embeddings to the expected output feature dimensionality.
            (:obj:`None`, :obj:`"last"`, :obj:`"cat"`, :obj:`"max"`,
            :obj:`"lstm"`). (default: :obj:`None`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.EdgeConv`.
    """
    supports_edge_weight = False
    supports_edge_attr = False

    def init_conv(self, in_channels: int, out_channels: int,
                  **kwargs) -> MessagePassing:
        mlp = MLP(
            [2 * in_channels, out_channels, out_channels],
            act=self.act,
            act_first=self.act_first,
            norm=self.norm,
            norm_kwargs=self.norm_kwargs,
        )
        return EdgeConv(mlp, **kwargs)



__all__ = ['GCN', 'GraphSAGE', 'GIN', 'GAT', 'PNA', 'EdgeCNN']


from torch_geometric.loader import DataLoader
from torch.utils.data import WeightedRandomSampler
import torch

def get_train_loader(train_dataset, batch_size, num_workers, seed):
    """
    Get train loader with over sampling to handle imbalanced dataset.
    Args:
        train_dataset: training dataset
        batch_size: batch size for training
        num_workers: number of workers for data loading
        seed: random seed for reproducibility
    Returns:
        train_loader: DataLoader with weighted random sampling
    
    """
    num_train_active = len(torch.nonzero(torch.tensor([data.y for data in train_dataset])))
    num_train_inactive = len(train_dataset) - num_train_active
    print(f'training # of molecules: {len(train_dataset)}, actives: {num_train_active}')

    train_sampler_weight = torch.tensor([(1. / num_train_inactive)
                                         if data.y == 0
                                         else (1. / num_train_active)
                                         for data in
                                         train_dataset])

    generator = torch.Generator()
    generator.manual_seed(seed)

    train_sampler = WeightedRandomSampler(weights=train_sampler_weight,
                                          num_samples=len(
                                          train_sampler_weight),
                                          generator=generator)

    train_loader = DataLoader(
        train_dataset,
        batch_size=batch_size,
        sampler=train_sampler,
        num_workers=num_workers,
        pin_memory=False
    )

    return train_loader

def get_test_loader(test_dataset, batch_size, num_workers):
    """
    Get test loader.
    Args:
        test_dataset: test dataset
        batch_size: batch size for testing
        num_workers: number of workers for data loading
    """
    num_test_active = len(torch.nonzero(torch.tensor([data.y for data in test_dataset])))
    print(f'test # of molecules: {len(test_dataset)}, actives: {num_test_active}')

    test_loader = DataLoader(
        test_dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        pin_memory=True
    )

    return test_loader

def get_valid_loader(valid_dataset, batch_size, num_workers):
    """
    Get validation loader.
    Args:
        valid_dataset: validation dataset
        batch_size: batch size for validation
        num_workers: number of workers for data loading
    """
    num_valid_active = len(torch.nonzero(torch.tensor([data.y for data in valid_dataset])))
    print(f'validation # of molecules: {len(valid_dataset)}, actives: {num_valid_active}')

    valid_loader = DataLoader(
        valid_dataset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        pin_memory=True
    )

    return valid_loader